#PROJECT : VNC 3.3 client and server for Inferno
#AUTHOR :  Marc Hufschmitt, marcas1@mime.univ-paris8.fr

# The main module for VNC client 

implement VNCclient;

# Client serveur VNC pour Inferno
# ########################

DEBUGLEVEL:=0;

include "sys.m";
 sys: Sys;
 stderr: ref Sys->FD;

include "tk.m";
tk : Tk;
Toplevel : import tk;

include "string.m";
str : String;


#include "sh.m";
#     sh : Command;

include "draw.m";
 draw: Draw;
Font, Point, Rect, Screen, Display, Image, Context: import draw;
#ctxt: ref Context;

include "rfbproto.m";
        rfb : RFBprotocol;
Rectangle, AuthMsg, VersionMsg, ClientInitialisation : import rfb;

include "vncauth.m";
vncauth : VNCauth;

include "VNCinclude.b";

include "csget.m";
csget : CsGet;

include "wmlib.m";
	wmlib: Wmlib;

VNCclient : module {
    init : fn(ctxt : ref Draw->Context, argv : list of string);
};

bigEndianPix:=int 1;
 framnreq:int;
# taille de la fenetre du client VNC
VNCviewerrect : Rect;
# taille de la partie visualisation de la fenetre
VNCviewrect : Rect;
RFBreq :=0;
RFBresp :=0;
# image ayant la taille et le contenu de l'ecran du serveur VNC
VNCimage : ref Image;


VNCservername : string;
VNCdisplay : int;
VNCview : ref Image;
VNCenableclip :=0;
VNCcliptext := "";
VNCcolormap := array[256] of int;
VNCbigendianflag :=0;
Infernocolormap : array of int;
VNCbitperpixel : int;

init(ctxt : ref Draw->Context, argv : list of string)
{
    # Chargemnt du module systeme
    sys = load Sys Sys->PATH;
    # en cas d'erreur fin sans message
    if (sys==nil) exit;
    # on definit le descripteur de sortie des erreurs
    stderr = sys->fildes(2);

    # Chargement des services de connection
    #if ((csget = load CsGet CsGet->PATH) == nil)
#	fatal(sys->fprint(stderr,"Impossible d'obtenir les services reseau\n"));
    #csget->init(ctxt,argv);
    # On cree un nouveau groupe de processus
    sys->pctl(sys->NEWPGRP, nil);

    #if (sys->bind("#I","/net",sys->MAFTER)<0)
#	fatal(sys->fprint(stderr,"Impossible d'obtenir les services reseau\n"));


    draw = load Draw Draw->PATH;
    if (draw==nil) 
	fatal(sys->fprint(stderr,"Chargement module Draw : %r\n"));
    if (ctxt.display.image.ldepth!=3)
	fatal(sys->fprint(stderr,"VNCviewer ne gere que les ecrans 8bits\n"));
    Infernocolormap = array[4096] of int;
    SetInfernocolormap(ctxt);

    wmlib = load Wmlib Wmlib->PATH;

    tk = load Tk Tk->PATH;
    if (tk==nil) 
	fatal(sys->fprint(stderr,"module Tk :%r\n"));

    str = load String String->PATH;

    rfb = load RFBprotocol RFBprotocol->PATH;
    if (rfb==nil) 
	fatal(sys->fprint(stderr,"Chargement module RFBprotocole : %r\n"));

    rfb->init();

    rfb->DEBUG=0;
    

      mfd := sys->open("/dev/pointer", sys->OREAD);
#      if(mfd == nil) {
#      sys->print("open: /dev/pointer: %r\n");
#      return;
#      }
      keyfd := sys->open("/dev/keyboard", sys->OREAD);
#      if(keyfd == nil) {
#      sys->print("can't open /dev/keyboard: %r");
#      return;
#      }
    

      appname: string;
    # On prend le nom de l'application
    if (argv!=nil) {
	appname = hd argv;
	argv = tl argv;
    }
    else {
	sys->fprint(stderr,"VNCclient : arguments manquants\n");
	return;
    }
  servername: string;
    # On prend le nom du serveur
    if (argv!=nil) {
	servername = hd argv;
	argv = tl argv;
    }
    else {
	sys->fprint(stderr,"%s : arguments manquants (serveur VNC)\n",appname);
	return;
    }


    port :int;
    if (argv!=nil) {
	port = int hd argv;
	argv = tl argv;
    }
    else {
	sys->fprint(stderr,"%s : arguments manquants (No de display): default 0\n",appname);
	port=0;
    }
    # valeur inferieure a 100, c'est un numero de display
    if (port < 100) port += 5900;
    VNCdisplay=port-5900;

    password :="";
    if (argv!=nil) {
	password = hd argv;
	argv = tl argv;
    }
     
    (ok,conn) := connecte(servername,port);
    if (ok<0) {
	fatal(sys->fprint(stderr,"connection impossible\n"));
	return;
    }
    c := ref conn;

    ctxt=initdisplay(ctxt);
    VNCrect := ctxt.screen.image.r;
    sys->print("(%d,%d [%d,%d])\n",VNCrect.min.x,VNCrect.min.y,VNCrect.dx(),VNCrect.dy());


    pvc :=ref VersionMsg;
    pvc.init();
    sys->print("%d %d\n",len pvc.struct,pvc.SIZE);
    if (rfb->ReadExact(c, pvc.struct, pvc.SIZE)<0) 
	fatal(sys->fprint(stderr,"Lecture Version %r\n")); 

    

    pvs:=ref VersionMsg;
    pvs.init();
    # on renvoie notre version du proto
    if (rfb->WriteExact(c,pvs.struct, pvs.SIZE)<0) 
	fatal(sys->fprint(stderr,"Ecriture Version %r\n")); 
    
    
    (dmaj,dmin):=pvc.cmp(pvs);
    sys->print("Difference des versions M%d m%d\n",dmaj,dmin);
    if (dmaj!=0 && dmin!=0)
	fatal(sys->fprint(stderr,"Version du protocole differentes\n"));

    pvs=nil;
    pvc=nil;
    

    auth := ref AuthMsg;
    #rfbNoAuth : con 1;
    auth.init(rfb->NoAuth);
    if (rfb->ReadExact(c, auth.struct, len auth.struct) < 0) 
	fatal(sys->fprint(stderr,"Lecture Auth %r\n")); 

    aut :=rfb->INT(auth.struct);
    case aut {

	# La connection a echoue
	rfb->ConnectionFailed =>
	reason :=array[200] of byte;
	sys->fprint(stderr,"Server send : connection failed\n");
	if (rfb->ReadExact(c, auth.struct, len auth.struct-1) < 0) 
	    fatal(sys->fprint(stderr,"Lecture raison de la non connection\n"));
	if (rfb->ReadExact(c, reason, rfb->INT(auth.struct)) < 0) 
	    fatal(sys->fprint(stderr,"Lecture raison de la non connection\n"));
	fatal(sys->fprint(stderr,"Serveur reason : %s\n",string reason[0:rfb->INT(auth.struct)]));
	
	# L'authentification VNC est demande
	rfb->VNCAuth =>
	    sys->print("VNC authentification\n");
	vncauth = load VNCauth VNCauth->PATH;
	if (vncauth == nil)
	    fatal(sys->fprint(stderr,"Impossible de charger"+VNCauth->PATH));

	# on lit le message aleatoire envoye par le serveur
	if (rfb->ReadExact(c, auth.keys, len auth.keys) < 0) 
	    fatal(sys->fprint(stderr,"Lecture Challenge Key%r\n")); 
	#sys->print("Challenge Key: %d\n",rfb->INT(auth[0:2]));
	#auth.keys = array[16] of {byte 16r15, byte 16rad, byte 16r0, byte 16re3, byte 16r16, byte 16r97, byte 16rb4, byte 16raf, byte 16r3a, byte 16rcf, byte 16r5f, byte 16rf2, byte 16re5, byte 16rb7, byte 16r2e, byte 16r36 };
	# on encrypte le message avec le mot de passe
	vncauth->EncryptBytes(auth.keys,password);
	sys->print("srv encry %s\n",rfb->hexstrbuf(auth.keys));

	if (rfb->WriteExact(c, auth.keys, len auth.keys) < 0) 
	    fatal(sys->fprint(stderr,"DES response%r\n")); 

	if (rfb->ReadExact(c, auth.struct, len auth.struct) < 0) 
	    fatal(sys->fprint(stderr,"Lecture auth result%r\n")); 

	sys->print("Authentification information: %d\n",rfb->INT(auth.struct));
	case rfb->INT(auth.struct) {
	    rfb->AuthOK =>
		sys->print("Authentification reussie\n");
	    rfb->Authfailed =>
		sys->print("Authentification ratee\n");
	    return;
	    rfb->Authtoomany =>
		sys->print("Authentifications rastees trop nombreuses : client banni\n");
	    return;
	    # cas non repertories
	    * =>
		fatal(sys->fprint(stderr,"Server send : unknown authentification method\n"));
	}
	rfb->NoAuth => sys->print("Pas d'authentification\n");
	* => fatal(sys->fprint(stderr,"Auth unknown %d\n",aut));
    }
    
    cim := ref ClientInitialisation;
    if (cim.write(c,cim.Shared)<0)
	fatal(sys->fprint(stderr,"%s",rfb->lasterror()));
    
    ServerMsg :=array[24] of byte;
    if (rfb->ReadExact(c, ServerMsg, len ServerMsg) < 0) 
	fatal(sys->fprint(stderr,"Serveur initialisation%r\n")); 
    
    largeur :=rfb->INT(ServerMsg[0:2]);
    hauteur :=rfb->INT(ServerMsg[2:4]);
    namelen :=rfb->INT(ServerMsg[20:24]);

    VNCbitperpixel =rfb->INT(ServerMsg[4:5]);
    depth :=rfb->INT(ServerMsg[5:6]);
    VNCbigendianflag =rfb->INT(ServerMsg[6:7]);
    
    truecolourflag :=rfb->INT(ServerMsg[7:8]);

    #colormap : array of int;
    colormode := rfb->PixFormatNone;
    if (truecolourflag) {
	redmax :=rfb->INT(ServerMsg[8:10]);
	greenmax :=rfb->INT(ServerMsg[10:12]);
	bluemax :=rfb->INT(ServerMsg[12:14]);
	
	redshift :=rfb->INT(ServerMsg[14:15]);
	greenshift :=rfb->INT(ServerMsg[15:16]);
	blueshift :=rfb->INT(ServerMsg[16:17]);
	sys->print("(%d %d %d) (%d %d %d) bigendian[%d] truecolor[%d]\n",redmax, greenmax, bluemax, redshift, greenshift, blueshift,VNCbigendianflag,truecolourflag);
	if (redmax==16r1f && greenmax==16r3f && bluemax==16r1f &&
	    redshift==16rb && greenshift==16r5 && blueshift==16r0)
	    colormode=rfb->PixFormat16bits;
	if (redmax==16rff && greenmax==16rff && bluemax==16rff &&
	    redshift==16r10 && greenshift==16r8 && blueshift==16r0)
	    colormode=rfb->PixFormat24bits;
	if (redmax==16r7 && greenmax==16r7 && bluemax==16r3 &&
	    redshift==16r0 && greenshift==16r3 && blueshift==16r6)
	    colormode=rfb->PixFormatBGR233;
	if (colormode==rfb->PixFormatNone)
	    fatal(sys->fprint(stderr,"Format de pixels non gere %s\n",rfb->hexstrbuf(ServerMsg)));
	
	#  colormap = array[1<<bitperpixel] of int;
#  	    sys->print("%d %d,",len colormap,1<<(bitperpixel-1));
#  	for (i:=0; i<len colormap; i++) {
#  	    if ((i+1)%16r800==0)
#  		sys->print("%d = %d\n",i,((i>>redshift) & redmax)*255/(redmax));
#  	    colormap[i] = 
#  		ctxt.display.rgb2cmap(((i>>redshift) & redmax)*255/redmax,
#  				      ((i>>greenshift) & greenmax)*255/greenmax,
#  				      ((i>>blueshift) & bluemax)*255/bluemax
#  				      );
    }
    else {
	colormode=rfb->PixFormatCMap;
    }
    
    
    
    sys->print("Server Init\n%s\n",rfb->hexstrbuf(ServerMsg[4:17]));
    
    
    servname := array[namelen] of byte;

    if (rfb->ReadExact(c, servname, namelen) < 0) 
	fatal(sys->fprint(stderr,"server name%r\n")); 
    
    sys->print("Serveur %s: [%dx%d]\n",VNCservername=string servname[0:namelen],largeur,hauteur);
    VNCimage= ref Image;

    VNCimage=ctxt.display.newimage(((0,0),(largeur,hauteur)),ctxt.screen.image.ldepth,0,Draw->Black);

    spawn clientwindow(ctxt,c);
    
    if (1) {
	ServerMsg[0]=byte 0; # type du message
	ServerMsg[6]= byte 1;  # Force big endian
	if (rfb->WriteExact(c, ServerMsg, 20) < 0) 
	    fatal(sys->fprint(stderr,"SetPixelFormat%r\n")); 
    }

    if (1) {
	se :=array[24] of byte;
	se[0:]=rfb->CARD(8,2);
	se[2:]=rfb->CARD(16,5); # 5 type
	se[4:]=rfb->CARD(32,0);
	se[8:]=rfb->CARD(32,1);
	se[12:]=rfb->CARD(32,2);
	se[16:]=rfb->CARD(32,4);
	se[20:]=rfb->CARD(32,5);
	if (rfb->WriteExact(c, se, 24) < 0) #24
	    fatal(sys->fprint(stderr,"SetEncodings%r\n")); 
    }

    
    msg := array[1] of byte;

    n :int;
    framnreq=0;
    

    #spawn mouse(c, ctxt.screen, mfd);
    #spawn keyboard(c, ctxt.screen, keyfd);
    #spawn dispmouse(chanmouse);
    while(1)
    {
	if (rfb->Read_Data(c,msg)<0) {
	    sys->fprint(stderr,"Le serveur s'est deconnecte\n");
	    return; 
	}
	case int msg[0] {
	    rfb->mtFramebufferUpdate=>

	    
	    fbu :=array[4] of byte;
	    if (rfb->ReadExact(c, fbu[1:], len fbu[1:]) < 0) 
		fatal(sys->fprint(stderr,"initial Framebuffer update%r\n"));
	    
	    numrect :=rfb->INT(fbu[2:4]);
	    #sys->fprint(stderr,"Reception de %d rectangles\n",numrect);
	    for (nr:=0; nr < numrect; nr++) {
	      rect:=array[12] of byte;
		if (rfb->ReadExact(c, rect, len rect) < 0) 
		    fatal(sys->fprint(stderr,"initial Framebuffer request%r\n")); 
		
		rx :=rfb->INT(rect[0:2]);
		ry :=rfb->INT(rect[2:4]);
		rl :=rfb->INT(rect[4:6]);
		rh :=rfb->INT(rect[6:8]);
		re :=rfb->INT(rect[8:12]);
		#sys->fprint(stderr,"rect n%d : x%dy%d [%dx%d] en %d\n",nr,rx,ry,rl,rh,re);
		#sys->print("%d ", re);
		rectdst :=Draw->Rect((rx,ry),(rx+rl,ry+rh));
		#byteperpixel := bitperpixel/8;
		#sys->print("%d ", VNCbitperpixel);
		case re {
		    # Mode Raw
		    
		    rfb->Encodingraw =>

		    buf :=array[rl*rh*(VNCbitperpixel/8)] of byte;
		    if (rfb->ReadExact(c, buf, len buf) < 0) 
			fatal(sys->fprint(stderr,"rect raw%r\n")); 
		    pixels := array[rl*rh] of byte;
		    rfbpix2cmap(ctxt,buf,pixels,colormode);
		    #pixels = rfb->ImageColorFromPixelFormat(t.image.display,buf,colormode,colormap,bigendianflag);
		    		    
		    VNCimage.writepixels(rectdst,pixels);
		    
		    
		    #imagetmp :=ctxt.display.newimage(rectdst,3,0,0);
		    #if (imagetmp==nil)
		    #    fatal(sys->fprint(stderr,"Pas d'image t'es pas sage\n"));
		    #imagetmp.writepixels(rectdst,buf);
		    #VNCimage.draw(rectdst,imagetmp,nil,imagetmp.r.min);
		    
		    #t.image.writepixels(VNCviewrect.addpt(rectdst.min),pixels);
		    #updatechan <-= rectdst;
		    RFBresp++;
		    
		    # COPY RECTANGLE ok
		rfb->Encodingcopyrect =>
		    
		    srcpt :=array[4] of byte;
		    if (rfb->ReadExact(c, srcpt, len srcpt) < 0) 
			fatal(sys->fprint(stderr,"rect copy rectangle %r\n"));

		    x :=rfb->INT(srcpt[0:2]);
		    y :=rfb->INT(srcpt[2:4]);
		    buf :=array[rl*rh*(VNCbitperpixel/8)] of byte;
		    rectsrc :=Draw->Rect((x,y),(x+rl,y+rh));

		    VNCimage.readpixels(rectsrc,buf);
		    VNCimage.writepixels(rectdst,buf);
		    
		    # La modification est le rectangle contenant src et dst
		    #modif := ORrect(rectsrc,rectdst);
		    #updatechan <- = modif;
		    RFBresp++;
		    
		    # RRE and coRRE encoding 
		    rfb->EncodingRRE or rfb->EncodingCoRRE => 
		    # Meme entete de base
		    buf :=array [4+VNCbitperpixel/8] of byte;
		    if (rfb->ReadExact(c, buf, len buf) < 0) 
			fatal(sys->fprint(stderr,"rect (co)RRE %r\n"));
		    
		    
		    RREstruct : array of byte;
		    if (re==rfb->EncodingRRE)
			RREstruct=array[VNCbitperpixel/8+8] of byte;
		    else # rfb->EncodingCoRRE
			RREstruct=array[VNCbitperpixel/8+4] of byte;

		    RREimage :=ctxt.display.newimage(((0,0),(1,1)),
						     VNCimage.ldepth,
						     1,0); # repl=1 enlarge
		    # un rectangle taille serveur aux bornes inversees
		    #modif := Rect (VNCimage.r.max,VNCimage.r.min);
		    
		    # Rectangle maitre
		    couleurs := array[1] of byte;
		    #couleurs[0] = byte rfb->ColorFromPixelFormat(t.image.display,buf[4:4+byteperpixel],colormode,colormap,bigendianflag);
		    rfbpix2cmap(ctxt,buf[4:4+VNCbitperpixel/8],couleurs,colormode);
		    # on met le point de couleur 
		    RREimage.writepixels(RREimage.r,couleurs);
		    # on ecrit ce pave de couleur sur l'image VNC
		    VNCimage.draw(rectdst,RREimage,nil,RREimage.r.min);

		    # traitement des sous-rectangles 
		    nsubr:=rfb->INT(buf[0:4]);
		    for (i:=0; i<nsubr; i++) {
			if (rfb->ReadExact(c, RREstruct, len RREstruct) < 0) 
			    fatal(sys->fprint(stderr,"read RREstruct %r\n"));
			#couleurs[0] = rfb->ColorFromPixelFormat(t.image.display,RREstruct[0:byteperpixel],colormode,colormap,bigendianflag);
			rfbpix2cmap(ctxt,RREstruct[0:VNCbitperpixel/8],couleurs,colormode);
			coo := RREstruct[VNCbitperpixel/8:];
			r : Draw->Rect;
			if (re==rfb->EncodingRRE) {
			    r = ((rfb->INT(coo[0:2]),
				  rfb->INT(coo[2:4])),
				 (rfb->INT(coo[0:2])+rfb->INT(coo[4:6]),
				  rfb->INT(coo[2:4])+rfb->INT(coo[6:8])));
			}
			else {
			    r = ((rfb->INT(coo[0:1]),
				  rfb->INT(coo[1:2])),
				 (rfb->INT(coo[0:1])+rfb->INT(coo[3:6]),
				  rfb->INT(coo[1:2])+rfb->INT(coo[3:4])));
			}
			
                        # Le sous rectangle est relative au rectangle maitre
			r.addpt(rectdst.min);
			# On coupe tout ce qui depasse...oups
			r.Xrect(rectdst);
			
			
			RREimage.writepixels(RREimage.r,couleurs);
			VNCimage.draw(r,RREimage,nil,RREimage.r.min);
			if ( i%((nsubr/10)+1)) {
			    #sys->print("RRE %s %s\n",printrect(r),printrect(modif)); 
			    #updatechan <- = modif;

			}

			RFBresp++;
			
		    }
		    
		    # hextile Encoding !!!
		    rfb->Encodinghextile => 
		    sem := array[1] of byte;
		    if (rfb->ReadExact(c, sem, len sem) < 0) 
			fatal(sys->fprint(stderr,"read sub encoding mask %r\n"));
		    if (rfb->INT(sem[0:1]) & rfb->hextileRaw) {
			sys->print("hextile Raw");
			pixels := array [16 * 16*(VNCbitperpixel/8)] of byte;
			if (rfb->ReadExact(c, pixels, len pixels) < 0) 
			    fatal(sys->fprint(stderr,"read rawtile %r\n"));
			
		    }
		    else {
			if (rfb->INT(sem[0:1]) & 2) {
			    sys->print("hextile backspec");
			}
			if (rfb->INT(sem[0:1]) & 4) {
			    sys->print("hextile forespec");
			}
			if (rfb->INT(sem[0:1]) & 8) {
			    sys->print("hextile anysub");
			}
			if (rfb->INT(sem[0:1]) & 16) {
			    sys->print("hextile subrectcol");
			}
		    }
		    
		    * =>
			fatal(sys->fprint(stderr,"Encoding unknown %d\n",re));
		}
	    }
	    
	    
	    
	    rfb->ServerCutText=>
	    ct :=array[8] of byte;
	    if (rfb->ReadExact(c, ct[1:], len ct -1) < 0) 
		fatal(sys->fprint(stderr,"server Cut text request%r\n"));
	    
	    textlen :=rfb->INT(ct[4:8]);
	    if (textlen>0)
	      text:=array[textlen] of byte;
	    if (rfb->ReadExact(c, text, textlen) < 0) 
		fatal(sys->fprint(stderr,"cut text content request%r\n")); 

	    VNCcliptext = string text[0:textlen];
	    #sys->print("Le serveur a copie : \"%s\"\n",string text[0:textlen]);


	    
            # charge la palette de couleurs
	    rfb->mtSetColourMapEntries=>
	    sys->fprint(stderr,"Setcolourmapentries\n");
	    if (truecolourflag!=0) 
		fatal(sys->fprint(stderr,"Pas de table de couleur en true color\n"));
	    scme :=array[6] of byte;
	    if (rfb->ReadExact(c, scme[1:], len scme[1:]) < 0) 
		fatal(sys->fprint(stderr,"set color map entries%r\n")); 
	    #colormap = array [256] of int;
	    ncol :=rfb->INT(scme[4:6]);
	    fcol :=rfb->INT(scme[2:4]);
	    sys->fprint(stderr,"Setcolourmapentries %d->%d\n",fcol,ncol+fcol-1);
	    
	    for (nc:=fcol; nc< fcol+ncol; nc++) {
		if (rfb->ReadExact(c, scme, len scme) < 0) 
		    fatal(sys->fprint(stderr,"color[%d]%r\n",nc)); 
		# Les couleurs acceptes par rgb2cmap sont 8 bits (pas 16 bits)
		VNCcolormap[nc]=ctxt.display.rgb2cmap(rfb->INT(scme[0:2])/256,
						      rfb->INT(scme[2:4])/256,
						      rfb->INT(scme[4:6])/256);
		
  		#sys->print("couleur %d (%d,%d,%d)=%d\n",nc,rfb->INT(scme[0:2]),		rfb->INT(scme[2:4]),		rfb->INT(scme[4:6]),		colormap[nc]);
		       
	    }
	    rfb->Bell =>
	    sys->fprint(stderr,"Beep\n");
	    
	}
    }
}
    
connecte(serv : string, port : int) : (int, sys->Connection)
{
  dialto:="tcp!"+serv+"!"+string port;
    (ok,conn):=sys->dial(dialto,nil);
  
    return (ok,conn);
}
    
updater(c: ref sys->Connection, ctxt : ref Draw->Context, top : ref Tk->Toplevel)
{
    moderator :=0;
    incremental:=0;
    # Nombre de tours de boucle standart et maxi avant requete buffer
    WAITMIN, WAITSTD, WAITMAX : con iota*30;
    TIMEMAX : con 5; # 5 ms pour le systeme, LI-be-rez-LES-res-sources !!

    sys->print("WAIT %d %d %d\n", WAITMIN, WAITSTD, WAITMAX);

    # Affichage d'un message initial
    font := Font.open(ctxt.display, "*default*");
    top.image.text(top.image.r.min.add((40,VNCviewrect.dy()/2)),
		   ctxt.display.color(Draw->Black), (0,0), 
		   font, "Chargement de l'image initiale...");

    #le client fait sa premiere demande (initial = ~ incremental)    
    rfb->Send_FramebufferUpdateRequest(c,0,VNCviewrect);    
    for (;;)
    {
	
	# Le moderateur fait passer le temps
	moderator++;
	
	# Attente WAITMAX atteinte ? 
	if (!(moderator%WAITMAX)) { 
	    #sys->print("ALL %d\n",moderator);
	    # on a du temps a perdre pour demander tout l'ecran


	    rfb->Send_FramebufferUpdateRequest(c,1,VNCimage.r);
	    #rfb->Send_FramebufferUpdateRequest(c,0,VNCviewrect);

	    #sys->print("Je m'ennuie\n"); 
	}

	# Si un rafraichissement est demande par le tk
	# 1. on a attendu patiemment : on demande la partie visible
	# 2. on a beaucoup de rafraichissements a demander : calmer le jeu
	if (RFBresp>0 ) { #&& (!moderator%WAITSTD)) {
	    #sys->print("RESP %d\n",RFBresp);
	    showVNCview(ctxt,top);
	    #sys->print("moderator %d\t",moderator);
	    rfb->Send_FramebufferUpdateRequest(c,1,VNCviewrect);	    
	    RFBresp--;
	    }
	
	if (RFBreq>0 ) { #&& !moderator%WAITSTD) {
	    #sys->print("REQ %d\n",RFBreq);


	    #VNCviewrect.Xrect(updrect))
	    showVNCview(ctxt,top);
	    # Suppression d'une demande de mise a jour envoyee
	    RFBreq--;;
	}

	sys->sleep(TIMEMAX);
    }
}

mouse(c : ref sys->Connection, s: ref Draw->Screen, fd: ref Sys->FD)
{
    n := 0;
    buf := array[100] of byte;
    for(;;) {
	n = sys->read(fd, buf, len buf);
	if(n <= 0)
	    break;
	
	if(int buf[0] == 'm' && n == 37) {
	  pe:=array[6] of byte;
	    pe[0]=byte 5;
	    pe[2:] = rfb->CARD(16,int(string buf[ 1:13]));
	    pe[4:] = rfb->CARD(16,int(string buf[12:25]));
	    pe[1:] = rfb->CARD(8,int(string buf[24:37]));
	    
	    if (rfb->WriteExact(c, pe, len pe) < 0) 
		fatal(sys->fprint(stderr,"Pointer Event%r\n")); 
	    
	    
	}
    }
}

keyboard(c : ref sys->Connection, s: ref Draw->Screen, dfd: ref Sys->FD)
{
    b := array[1] of byte;
    buf := array[10] of byte;
    i := 0;
    for(;;) {
	# lecture de 10-i caractere
	n := sys->read(dfd, buf[i:], len buf - i);
	if(n < 1){
	    sys->fprint(stderr,"keyboard read error: %r");
	    break;
	}
	i += n;
	while(i >0 && (nutf := sys->utfbytes(buf, i)) > 0){
	    str := string buf[0:nutf];
	    #sys->print("Clavier %d:%s %s\n", nutf, str[0:nutf], rfb->hexstrbuf(buf[0:nutf]));
	    for (j := 0; j < len str; j++) {
		
	      ke:=array[8] of byte;
		ke[0]=byte 4;
		ke[1] = byte 1; # down (just press)
		ke[4:] = rfb->CARD(32,int str[j]);
		if (rfb->WriteExact(c, ke, len ke) < 0) 
		    fatal(sys->fprint(stderr,"Key Event%r\n")); 
	    }
	    buf[0:] = buf[nutf:i];
	    i -= nutf;
	}
    }
}

dispmouse(c : chan of ref Draw->Pointer)
{
    while(1) {
	k := <- c;
	#sys->print("display keyboard %d\n",k.xy.x);
    }
}

clientwindow(ctxt : ref Context, conn : ref Sys->Connection)
{
    recran :=ctxt.display.image.r;
    VNCviewerrect = ((recran.dx()/8,recran.dy()/8),(7*recran.dx()/8+1,7*recran.dy()/8+1));

  #scanvas:string;
    viewer_cfg := array[] of {
	"menu .m",
	".m add  cascade -label {Encodage} -menu .m.enc",
	".m add cascade -label {Presse papier} -menu .m.clp",

	".m add command -text Quitter -command {send mctxt quit}",
	
	"menu .m.enc",
	".m.enc add radiobutton -text {Raw} -command {send mctxt encode raw}",
	".m.enc add radiobutton -text {RRE} -command {send mctxt encode rre}",
	".m.enc add radiobutton -text {coRRE} -command {send mctxt encode corre}",
	".m.enc add radiobutton -text {hextile} -command {send mctxt encode hextile}",
	".m.enc select",
	"menu .m.clp",
	".m.clp add checkbutton -text {Accepter Copier} -command {send mctxt srvcut}",
        ".m.clp add command -text Coller -command {send mctxt paste}",
        ".m.clp add command -text Envoyer -command {send mctxt send}",

        sys->sprint("frame .ft -width %d -height %d -background red",VNCviewerrect.dx(),VNCviewerrect.dy()),
	#"frame .ft", # -width %d -heigth %d",
	
	# les xyscrollcommand mettent a jour les barres de scroll
	sys->sprint("canvas .ft.im -width %d -height %d -yscrollcommand {.ft.sy set} -xscrollcommand {.ft.sx set}"
		    ,VNCimage.r.dx(),VNCimage.r.dy()),
	    "bitmap .ft.fr0 -bitmap @/lib/vnc.bit",
        ".ft.im create bitmap 0 0 -bitmap .ft.fr0 -anchor nw",
	    
	    
	    #"scrollbar .ft.sy -command {.ft.im yview}",
	    #"scrollbar .ft.sx -orient horizontal -command {.ft.im xview}",
	    "scrollbar .ft.sy -command {send scroll y}",
	    "scrollbar .ft.sx -orient horizontal -command {send scroll x}",
	    
	    "pack .ft.sy -side right -fill y",# -in .ft",
	    "pack .ft.sx -side bottom -fill x",# -in .ft",
	    "pack .ft.fr0 -fill both -expand 1 -in .ft.im",
	    "pack .ft.im",
	    
	    #"pack .ft -fill both",
	    "pack propagate . 0",
	    "focus .ft.im",
	    
	    #"bind  .ft.im <ButtonPress> {send edit hello %x %y %b}",
	    "bind .ft.im <ButtonPress> {send edit mousep %x %y %b}",
	    "bind .ft.im <ButtonRelease> {send edit mouser %x %y %b}",
	    "bind .ft.im <Double-ButtonPress-1> {send edit moused %x %y %b}",
	    "bind .ft.im <Motion> {send edit mousem %x %y %b}",


	    
	    "bind .ft.im <KeyPress> {send edit clavier 1 %K}",
	    "bind .ft.im <KeyRelease> {send edit clavier 0 %K}",
	    # le menu apparait sur click droit sur la barre de titre
	    "bind .Wm_t.title <ButtonPress-2> {send edit clipbd %X %Y}",
	    #"bind .Wm_t.title <ButtonPress-2> {send edit shmu}",
	    # le choix du menu est fait losque le button est relache
	    "bind .m <ButtonRelease> {.m tkMenuButtonUp %x %y}",
	    "bind .m.enc <ButtonRelease> {.m.clp tkMenuButtonUp %x %y}",
	    "bind .m.clp <ButtonRelease> {.m.clp tkMenuButtonUp %x %y}",
	    "bind .ft.sy <ButtonRelease-1> + {send scroll y moveto final}", 
        #{ %W tkScrolBut1R; %W activate %x %y ;send scroll y moveto final}",
	    "bind .ft.sx <ButtonRelease-1> + {send scroll x moveto final}",
	    
	    "update"

	};


    
    wmlib->init();
    tkargs := sys->sprint(" -x %d -y %d",
			  VNCviewerrect.min.x,
			  VNCviewerrect.min.y);
    
    (t, titlectl) := wmlib->titlebar(ctxt.screen, 
				    tkargs, "VNC viewer : "+VNCservername+
				    ":"+string VNCdisplay,
				    wmlib->Appl|wmlib->Help);
    # Affichage de la fenetre
    wmlib->tkcmds(t, viewer_cfg);
    # Reglage de la taille
    tk->cmd(t,". configure -width "+string VNCviewerrect.dx()+" -height "+string VNCviewerrect.dy()+"; pack .ft ; update");
    
    VNCimage.text((0,0),ctxt.display.ones,(0,0),nil,"Loading");
    spawn updater(conn,ctxt,t);
    VNCviewrect=updateview(t);    
    eventchan := chan of string;
    tk->namechan(t, eventchan, "edit");
    scrolchan := chan of string;
    tk->namechan(t, scrolchan, "scroll");
    menucontext := chan of string;
    tk->namechan(t, menucontext, "mctxt");
    tk->cmd(t,"update");

    #menuindex : string;
    movetomsg :="";
  #moderate:=0;
    #t.image.flush(Draw->Flushoff);
    for(;;) 
	alt {
	    
	    menu := <-titlectl =>
	    
	    #sys->fprint(stderr,"menu %s\n",menu);
	    wmlib->titlectl(t, menu);
	    
	    
	    sys->print("%s\n",menu);
	    
	    case menu {
		"exit" => fatal(sys->print("Fin du client VNC\n"));
		
		"size" => 
		# forcer l'image .ft.im a synchroniser les ascenseurs
		tk->cmd(t,".ft.im xview scroll 0 unit; "+
			".ft.im yview scroll 0 unit"+
			"; update");
		# La vue a change de taille
		VNCviewrect = updateview(t);
		# On affiche le nouveau rectangle affichable
		RFBreq++;
		
		
		"move" =>
		# recuperer le focus sur la fenetre de visualisation
		tk->cmd(t,"focus .ft.im ; update");
		#sys->print("move\n");
		RFBreq++;
		
		* => sys->print("controle inconnu %s\n",menu);
		    
	    }

	    # Gestion du menu
	    menuctxt := <- menucontext =>
	    (nil,l):=sys->tokenize(menuctxt," ");
	    mct:= hd l;
	    case mct {
		* => sys->print("Controle menu no implemente\n");
		"srvcut" =>
		# On accepte/refuse de prendre le presse papier du client
		VNCenableclip ^=1;
		if (VNCenableclip) {
		    wmlib->snarfput(string VNCcliptext);
		}
		
		# On envoie le presse papier gere par le window manager
		"paste" =>
		#menuindex="0";
		rfb->Send_ClientCutText(conn,wmlib->snarfget());
		sys->print("Le client envoie son presse papier\n");
		
		# On envoie le presse papier 
		# comme s'il avait ete saisi au clavier
		"send" =>
		texte :=wmlib->snarfget();
		for (i:=0; i<len texte; i++)
		    rfb->Send_KeyEvent(conn,texte[i],rfb->KeyDown);
		
		"quit" =>
		#sys->fprint(stderr,"->%s\n",
			#    
		sys->fprint(stderr,"%s\n",tk->cmd(t,"label .ft.fr0 -bitmap @/lib/vnc.bit"));
		#sys->fprint(stderr,"%s\n",tk->cmd(t,".ft. create bitmap 0 0 -bitmap @/usr/marcas1/vnc/vnc.bit -anchor nw"));

		#sys->fprint(stderr,"%s\n",tk->cmd(t," winfo children .ft.im  "));
		sys->fprint(stderr,"%s\n",tk->cmd(t,".ft.fr0 cget -bgimage"));
		sys->fprint(stderr,"%s imageput\n",tk->imageput(t,".ft.fr0 configure -bgimage",ctxt.screen.image,nil));
		(imi,imm,err):=tk->imageget(t,".ft.fr0");
		if (imi!=nil) 
		    #ctxt.screen.image.draw(ctxt.screen.image.r,imi,nil,(0,0));
		    sys->fprint(stderr,"(%s)\n",err);
		l:=tk->windows(ctxt.screen);
		sys->print("%d windows tk\n",len l);
		#lo := l;
                #fatal(sys->fprint(stderr,"Quitter\n"));

		"encode" =>
		case hd tl l {

		    "raw" => ;
		    "coRRE" =>;
		    "RRE" =>;
		    "hextile" =>;
		}
	    }
	    
	    #updrect := <- updatechan =>
	    #if (moderate>10 & VNCviewrect.Xrect(updrect))
	    #{
	#	moderate=0;
	#	showVNCview(ctxt);
	#    }else moderate++;

	    #sys->fprint(stderr,"update\n");
	    
#wmlib->imageput(t,"",t.image,t.image));#VNCimage,VNCimage));
#(i,m,e):=tk->imageget(t,".ft");
#sys->print("update\nVNCviewerrect: %s\nVNCviewrect:%s\nVNCserverrect:%s\nupdrect%s\n",printrect(VNCviewerrect),printrect(VNCviewrect),printrect(VNCserverrect),printrect(updrect));
#}
	    # gestion des barres d'ascenceur
	    # pour eviter les secousses
	    sc := <- scrolchan =>
	    #sys->fprint(stderr,"scroll %s\n",sc);
	    (nil,scrl) :=sys->tokenize(sc," ");
	    xy := hd scrl; scrl=tl scrl;
	    cmd := hd scrl; scrl=tl scrl; val:="";
	    while (scrl!=nil) { val = val+hd scrl+" "; scrl = tl scrl; }
	    if (cmd=="moveto" && val!="final ") {
		# Si l'utilisateur tire les ascenceurs on ne reaffiche pas,
		# on attend qu'il finisse
		movetomsg=sys->sprint(".ft.im %sview %s %s; update",xy,cmd,val);
	    }
	    else {
		if (cmd!="moveto") {
		    # La valeur des scrollbar est envoye a l'image
		    tk->cmd(t,sys->sprint(".ft.im %sview %s %s; update",xy,cmd,val));
		}
		else {
		    tk->cmd(t,movetomsg);
		    sys->print("Lache!!!!\n");
		    sys->print("fin moveto (%s)\n",movetomsg);
		    movetomsg="";
		}
		VNCviewrect=updateview(t);
		RFBreq++;
	    }

	    #showVNCview(ctxt);
	    	    
		comm := <- eventchan =>
	    #sys->fprint(stderr,"event %s\n",comm);
	    (np, l) := sys->tokenize(comm, " ");
	    commande := hd l; l= tl l;
	    #sys->fprint(stderr,"commande %s ",commande);	
	    if (np>0)
		case commande {
		    
		    # Evenement Souris
		    "mousep" or "mouser" or "mousem" or "moused" => 
		    
		    x :=int hd l; l = tl l;
		    y :=int hd l; l = tl l;
		    b :=int hd l;
		    if (commande[5]=='r') b=0; # bouton relache => aucun presse
		    b = ((1<<b)>>1) & 16rff;
		    #sys->print("%s %d %d %d\n",commande,x,y,b);
		    rfb->Send_PointerEvent(conn,x+VNCviewrect.min.x,y+VNCviewrect.min.y,int b);
		    if (commande[5]=='d') {
			rfb->Send_PointerEvent(conn,x+VNCviewrect.min.x,y+VNCviewrect.min.y,int 0);
			rfb->Send_PointerEvent(conn,x+VNCviewrect.min.x,y+VNCviewrect.min.y,int 1);
			
		    }

		    # Evenement presse papier
		    "clipbd" =>
		    x := int hd l - 50;
		    y := int hd tl l - int tk->cmd(t, ".m yposition 0") - 10;
		    tk->cmd(t, ".m activate 0; .m post "+string x+" "+string y+"; grab set .m; update");
		    
		    # Evenement clavier
		    "clavier" =>
		    down := int hd l; l = tl l;
		    touche := unhex(hd l);
		    
		    
		    rfb->Send_KeyEvent(conn,tkkey2Xkey(touche),down);
		    
		    * => sys->fprint(stderr,"TK %s unimplemented\n",commande);
		    
		}
	}
}
    # Les evenements provoquent une demande de mise a jour a updater()
    #framnreq++;
    

    



showVNCview(ctxt : ref Context, toplevel : ref Tk->Toplevel)
{
    
    if (VNCview == nil || !VNCviewrect.eq(VNCview.r))
	# on cree une image hors affichage de la partie visible de l'image VNC
	VNCview = ctxt.display.newimage(VNCviewrect, ctxt.screen.image.ldepth,0,draw->Black);

    #VNCview.draw(VNCimage.r,VNCimage,ctxt.display.ones,VNCimage.r.min);
    #tk->cmd(t, ".Wm_t.title configure -for(text {View: grabbed}");
    #r := imconfig(t,VNCview);
    #r := imconfig(t,VNCimage);
    r := widgposn(toplevel,".ft.im");
    r1 :=widgposn(toplevel,".ft");
    #sys->print("r %s r1 %s view %s\n",printrect(r),printrect(r1),printrect(VNCviewrect));
#ims := array[1] of {im};
    #mask := array [1] of ref Image;
    VNCimage.clipr=VNCviewrect;
    
    toplevel.image.draw(r,VNCimage, nil ,VNCviewrect.min);
    
    
}

#      im := ctxt.display.newimage(VNCimage.r, ctxt.screen.image.ldepth,0,draw->White);
#      im.draw(VNCimage.r,VNCimage,ctxt.display.ones,VNCimage.r.min);
#      #tk->cmd(t, ".Wm_t.title configure -text {View: grabbed}");
#      r := imconfig(t,im);
#      ims := array[1] of {im};
#      mask := array [1] of ref Image;
#      t.image.draw(r,ims[0], mask[0] ,ims[0].r.min);
    
    #updrect.addpt(t.image.r.max);
    #r :=VNCimage.r.subpt(t.image.r.min);
    # je copie dans l'image t.image
    # l'image VNCimage a partir de son point superieur gauche
    # Je masque avec des ones puisque je veux tout
    
    #t.image.draw(VNCimage.r,VNCimage,ctxt.display.ones,VNCimage.r.min);
    
#}

widgposn(t: ref Toplevel, name: string): Rect
{
  r : Rect;
  
    r.min.x = int tk->cmd(t, name + " cget -actx");
    r.min.y = int tk->cmd(t, name + " cget -acty");
  r.max.x = r.min.x + int tk->cmd(t, name + " cget -width");
  r.max.y = r.min.y + int tk->cmd(t, name + " cget -height");
  #r.max.x = r.min.x + int tk->cmd(t, ".ft.im cget -actwidth");
  #r.max.y = r.min.y + int tk->cmd(t, ".ft.im cget -actheight");
    
    return r;
}

imconfig(t: ref Toplevel, im: ref Draw->Image): Rect
{
    
    width := im.r.dx();
    height := im.r.dy();
    tr := widgposn(t, ".Wm_t");
    if(width < tr.dx())
	width = tr.dx();
    tk->cmd(t, ".ft configure -width " + string width
    + " -height " + string height + "; update");
    r := widgposn(t, ".ft");
    return widgposn(t, ".ft");
}


splitstring(str, sep : string) : list of string
{
    deb :=0;
    res : list of string;
    res = nil;
    s := str + sep[0:1];
    #sys->print(",%s,%s,\n",s,sep);
    for (i:=0; i<len s; i++) {
	for (j:=0; j < len sep; j++) {
	    #sys->print("%d %d %d\n",i,j,deb);
	    if (s[i]==sep[j] || i ==len s -1 ) {
		res = s[deb:i] :: res;
		#sys->print("trouve %s\n",s[deb:i]);
		deb = i+1;
	    }
	}
    }
    return res;
}


updateview(t : ref tk->Toplevel) : Rect
{
    (n,l):= sys->tokenize(tk->cmd(t,".ft.im xview")," ");
    xmin :=hd l; xmax := hd tl l;

    (n,l)= sys->tokenize(tk->cmd(t,".ft.im yview")," ");
    ymin := hd l; ymax := hd tl l;

    VNCviewrect := ((int (real xmin * real VNCimage.r.dx()),
		     int (real ymin * real VNCimage.r.dy())),
		    (int (real xmax * real VNCimage.r.dx()),
		     int (real ymax * real VNCimage.r.dy()))
		    );
    #sys->print("after scroll %s\n",printrect(VNCviewrect));
    return VNCviewrect;
}

unhex (s : string) : int
{
    v :=0;
    c : int;
    for (j:=0; j< len s; j++) {
	if (s[j]>='A' && s[j]<='F') 
	    c = s[j]-'A' +10;
	else if ( s[j]>='0' && s[j]<='9') 
	    c=s[j]-'0';
	else continue;
	v=v*16+c;
    }
    return v;
}



SetInfernocolormap(ctxt : ref Draw->Context)
{
    for (r:=0; r<16; r++)
	for (v:=0; v<16; v++)
	    for (b:=0; b<16; b++)
		Infernocolormap[(r<<8)|(v<<4)|b]=ctxt.display.rgb2cmap(r*16,v*16,b*16);
    
}

tkkey2Xkey(tkk : int ) :int
{
    touche : int;
    case tkk {
	16rE014 => touche = 16rFF51;
	16rE012 => touche = 16rFF52;
	16rE015 => touche = 16rFF53;
	16rE013 => touche = 16rFF54;
	16r007f => touche = 16rFFFF;
	16re016 to 16rE017 => touche = 16rFF9A + tkk - 16rE016;
	0 to 16r1f => touche = tkk | 16rFF00;
	16r20 to 16r7e => touche = tkk; # pas de changement ' '->127
	* => touche = int '?';
    }
    sys->print("%X touche %X\n",tkk,touche);
    return touche;
}

rfbpix2cmap(ctxt : ref Draw->Context, buf, pix : array of byte, mode : int)
{
    sp :=0;
    dp :=0;
    l := len buf;
    r,v,b : int;
    p : int;

    if (mode==rfb->PixFormatCMap) {
	while(sp<l) {
	    pix[dp]=byte VNCcolormap[int buf[sp]];
	    sp++; dp++;
	}
    }
    else {
	if (mode==rfb->PixFormat16bits) {
	    if (VNCbigendianflag) 
		while (sp<l) {
		    p =((int buf[sp+1]<<8) | (int buf[sp]));
		    (r,v,b)=(p>>8, p>>3, p<<3);
		    pix[dp]=byte Infernocolormap[((r & 16rf0)<<4) | (v & 16rf0) | ((b & 16rf0)>>4)];
		    sp+=2; dp++;
		}
	    else 
		while (sp<l) {
		    p= ((int buf[sp+1]) | ((int buf[sp])<<8));
		    (r,v,b)=(p>>8, p>>3, p<<3);
		    pix[dp]=byte Infernocolormap[((r & 16rf0)<<4) | (v & 16rf0) | ((b & 16rf0)>>4)];
		    sp+=2; dp++;
		}
	}
	
	else if (mode==rfb->PixFormat24bits) {
	    if (VNCbigendianflag)
		while (sp<l) {
		    (r,v,b)=(int buf[sp+2], int buf[sp+1] ,int buf[sp]);
		    pix[dp]=byte Infernocolormap[((r & 16rf0)<<4) | (v & 16rf0) | ((b & 16rf0)>>4)];
		    sp+=4; dp++;
		}
		else
		    while (sp<l) {
			(r,v,b)=(int buf[sp+1], int buf[sp+2], int buf[sp+3]);
			pix[dp]=byte Infernocolormap[((r & 16rf0)<<4) | (v & 16rf0) | ((b & 16rf0)>>4)];
			sp+=4; dp++;
		    }
	}
	
	else if (mode==rfb->PixFormatBGR233) {
	    while (sp<l) {
		p= int buf[sp];
		(r,v,b):=((p<<16r5) & 16re0,
			  (p<<16r2) & 16re0,
			  (p>>16r0) & 16rc0);
		pix[dp]=byte Infernocolormap[((r & 16rf0)<<4) | 
					     (v & 16rf0) | 
					     ((b & 16rf0)>>4)];
		sp++; dp++;
	    }
	}
	else
	    fatal(sys->fprint(stderr,"colormode inconnu\n"));
	
	#pix[dp]=byte ctxt.display.rgb2cmap((r & 16rf0),(v & 16rf0),(b & 16rf0));
	    #sys->print("%d->%d[%d]=%d ",sp,dp,((r & 16rf0)<<4) | (v & 16rf0) | ((b & 16rf0)>>4), int pix[dp]);	    
    }
}    
