#PROJECT : VNC 3.3 client and server for Inferno
#AUTHOR :  Marc Hufschmitt, marcas1@mime.univ-paris8.fr

# RFB protocol : Hextile encoding functions
# contient les fonctions necessaires a l'envoi en encodage hextile


# Envoie un rectangle au format hextile et retourne le nombre d'octets emis
SendRectEncodingHextile(c : ref Sys->Connection, s : ref Draw->Screen, r : rfb->Rectangle, pf : ref PixelFormat) : int
{
    bpp := int pf.PIXEL_FORMAT[0];

    # rafale envoi tous les pixel d'un coup
    rafale := 1;
    bufraf := array[4096] of byte;
    bufrafidx := 0;
    
    # Un tableau contenant l'encodage de chaque tile
    bufrect := array[(r.dy()/16+1)*(r.dx()/16+1)] of array of byte;
    # Un tableau contenant l'encodage en mode raw d'un tile
    bufout := array[16*16*(bpp/8)+1] of byte;
    # le contenu d'un tile sous Inferno
    bufscr := array[16*16] of byte;
    # le contenu d'un tile en pixel rfb
    validBG,validFG : int; validBG=validFG = 0;
    bg,fg : byte;
    Idx : int;
    # un tableau pour la conversion
    color := array[1] of byte;
    # Le nombre d'octets ecrits
    lonhex := 0;
    ntile := 0;
    
    for (y := r.min.y; y < r.max.y; y+=16) {
	
	for (x := r.min.x; x < r.max.x; x+=16) {
	    ntile++;
	    # On decompose en 16 par 16
	    w := 16;
	    h := 16;
	    # Si le rectangle est moins large que 16
	    if (r.max.x - x < 16)
		# on ajuste
		w = r.max.x - x;
	    # Si le rectangle est moins haut que 16
	    if (r.max.y - y < 16)
		# On ajuste
		h = r.max.y - y;
	    #sys->print("Tile %d/%d[%d,%d]\n",ntile-1,r.surf()/256,w,h);

	    # recommencons au debut du buffer
	    Idx = 1;
	    # On remet le hextile submask a 0
	    bufout[0]= byte 0;


	    hexrect := ((x,y),(x+w,y+h));
	    s.image.readpixels(hexrect,bufscr);
	    (mode,newbg,newfg) := DecomposeForHextile(bufscr,hexrect);
	    #sys->print("Mode %d\n", mode);

	    # premier tile ou nouveau fond on precise la couleur de fond
	    if ((!validBG) || (newbg!=bg)) {
		validBG = 1;
		bg = newbg;
		bufout[0] |= byte rfb->hextileBackgroundSpecified;
		color[0] = bg;
		cmap2rfbpix(color,bufout[Idx:],pf);
		Idx+=bpp/8;
	    }

	    # En mode solid on a fini le tile courant
	    if (mode == rfb->hextileModeSolid) {
		#bufrect[(y/16)*(1+s.image.r.dx()/16)+x/16] = bufout[0:Idx];
		#sys->print("buffer hextile #%d\n",(y/16)*(1+s.image.r.dx()/16)+x/16);
		#sys->print("Mode Solid %d\n",int bg);
	    }
	    else {
		# On va forcement avoir des sous rectangles de tile
		bufout[0] |= byte rfb->hextileAnySubrects;

		# en mode mono 
		if (mode==rfb->hextileModeMono) {
		    # premier tile ou nouvel av. plan  
		    # on precise la couleur av. plan
		    if ((!validFG) || newfg!=fg) {
			validFG = 1;
			fg = newfg;
			bufout[0] |= byte rfb->hextileForegroundSpecified;
			color[0] = byte newfg;
			cmap2rfbpix(color,bufout[Idx:],pf);
			Idx+=bpp/8;
		    }
		}
		else {
		    validFG=0;
		    # Multicolor : les sous rectangles ont des couleurs
		    bufout[0] |= byte rfb->hextileSubrectsColoured;
		}
		
		
		# on demande la decomposition du pave hextile et sa taille rfb
		(taille,ls) := subrectEncode(bufscr,(w,h),bg,mode,bpp);
		
		# Si l'encodage en sous rectangles est trop grand
		# on encode en raw
		if (taille>(w*h*bpp/8+1)) {
		    sys->print("vaut pas le coup %d %d\n",taille,(w*h*bpp/8+1));
		    Idx=1;
		    # on ecrase le sousmasque cree avant
		    bufout[0] = byte rfb->hextileRaw;
		    validBG = 0;
		    validFG = 0;
		    # on remet l'image dans bufscr
		    # puisque l'endage l'a corrompu
		    s.image.readpixels(hexrect,bufscr);
		    cmap2rfbpix(bufscr[0:w*h],bufout[Idx:],pf);
		    Idx+=w*h*bpp/8;
		}
		# sinon on va coder chaque sous-pave du tile courant
		else {
		    if ((len ls)==0)
			sys->print("subtile list empty");
		    # On donne le nombre de sous paves
		    bufout[Idx++]= byte len ls;
		    while(ls != nil) {
			sr := hd ls;
			#sys->print("subhextile %s total %d\n",printrect(sr.r),taille);
			if (mode!=rfb->hextileModeMono) {
			    # on a place la couleur dans le buffer
			    color[0] = sr.couleur;
			    cmap2rfbpix(color,bufout[Idx:],pf);
			    Idx+= bpp/8;
			}
			# On code les x-y et w-h
			bufout[Idx++]=byte (sr.r.min.x*16+sr.r.min.y);
			bufout[Idx++]=byte ((sr.r.dx()-1)*16+(sr.r.dy()-1));
			ls = tl ls;
		    }
		}
	    }
	    if (bufout[0]==byte 0 && Idx!=1) {
		sys->print("ERREUR\n");
		
		sys->print("hextilemode %d Idx %d dans hextile #%d\n",int bufout[0],Idx,(y/16)*(1+r.dx()/16)+x/16);
	    }

	    if (!rafale) {
		if (rfb->Send_Data(c,bufout[0:Idx])<0) {
		    c=nil;
		    debug(1,"Erreur ecriture : envoi rectangle en Hextile");
		    return 0;
		}
		#sys->sleep(10);
	    } else {
		# Si le buffer rafale est plein
		# ou que c'est le dernier tile
		bufraf[bufrafidx:] = bufout[0:Idx];
		bufrafidx += Idx;
		if ((bufrafidx > 4096-len bufout) || 
		    ((x+16 >= r.max.x) && (y+16 >= r.max.y))) {
		    if (rfb->Send_Data(c,bufraf[0:bufrafidx])<0) {
			c=nil;
			debug(1,"Erreur ecriture : envoi rectangle en Hextile");
			return -1;
			
		    }
		    lonhex+=bufrafidx;
		    bufrafidx = 0;
		}
		
		
	    }
	    
	} # next x +16
    } # nex y + 16
    
   
    return lonhex;
    
}
	    
subrectEncode(scr : array of byte, ptmax : draw->Point, bg : byte, mode : int,bpp : int) : (int, list of rfb->Pave)
{
    w := ptmax.x;
    h := ptmax.y;
	
    taille :=0;
    lp : list of rfb->Pave;
    hx, hy, vx, vy : int;
    vx = hx =0;

    # parcours des lignes
    for (y := 0 ; y < h ; y++) {
	#line = data+(y*w); 
        
	#parcours des pixels de la ligne
	for (x := 0; x < w ; x++) {
            
	    # Si le point est different du fond
	    if (scr[y*w+x] != bg) {
                # la couleur courante est celle du premier point
		cl := scr[y*w+x];
		# rectangle horizontal de hauteur ....
                hy := y;
		# rectangle horizontal a calculer
                hyflag := 1;
		
		# Pour chacune des lignes partant de la ligne y
                for (j := y; j < h; j++) {
                    #seg = data+(j*w);
		    
		    # si le point change de couleur ici on arrete
                    if (scr[j*w+x] != cl) {break;}

                    # i parcourt la ligne
		    i := x;
                    # tant que la couleur est la meme sur la ligne on avance
		    while ((i < w) && (scr[j*w+i] == cl)) 
			i++;

		    # si sur la premiere ligne la largeur 
		    # des deux rectangles potentiel s'arrete en i
                    if (j == y) vx = hx = i;
                    
		    # si le point de droite est plus petit que
		    # la fin du rectangle vertical on ajuste sa largeur max
		    if (i < vx) vx = i;                                       
                    
		    # si mode horizontal && i plus grand 
		    # que la fin du rectangle horizontal
		    if ((hyflag > 0) && (i >= hx)) {                          
                        # notre rectangle horizontal fini avant i
			hy += 1;                                              
                    } else {
			# sinon fin du rectangle horizontal
			hyflag = 0;
                    }                                                         
                }
                vy = j-1;
		# on choisi le plus grand des deux rectangles
		# rectangle horizontal (x,y,hx,hy) et        
		# rectangle vertical (x,y,vx,vy)
                 
		p1 : rfb->Pave;
		p2 : rfb->Pave;
		p1.r.min = (x,y);
		p2.r.min = (x,y);
		p1.r.max = (hx,hy);
		p2.r.max = (vx,vy);

                if ((p1.r.surf()) < (p2.r.surf())) {
		    p1.r = p2.r;
		}
		
		if (mode==rfb->hextileModeMono) {
		    # Il faudra 1 octet pour X-Y et un autre pour W-H;
		    taille+=2;
		}
		else {
		    p1.couleur = cl;
		    taille+=2+bpp/8;
		}
		
		#  Abandon si la taille depasse l'ecodage hextileRaw
		if (taille>(w*h*bpp/8+1))
		    return (taille,nil);


		lp = p1 :: lp;
		
		# Le rectangle est fait
		for (j = p1.r.min.y ; j < p1.r.max.y; j++) {
		    for (i := p1.r.min.x; i < p1.r.max.x; i++) {	  
			# On le marque a la couleur du fond
                        scr[j*w+i] = bg;
		    }
		}
	    }
	}
    }
    
    ltmp : list of rfb->Pave;
    # on retourne la liste
    while(lp!=nil) {
	ltmp = hd lp :: ltmp;
	lp = tl lp;
    }
    lp= ltmp;
    return (taille,lp); 
}

# retourne mode : 2 pour solid (1 couleur), 1 pour mono (2 couleurs), 0 sinon
# retourne bg = le pixel le plus frequent
# retourne fg le second pixel le + frequent
DecomposeForHextile(scr : array of byte, r : Rect) : (int, byte , byte)
{
    # mode 1 couleur (solid) et mode 2 couleur (mono) supposes vrais
    solid := 1;
    mono := 1;
    mode : int;

    # n1,n2 flags pour le cas du premier point
    n1 := 0;
    n2 := 0;
    col1, col2 : byte;

    fg,bg : byte;

    for (i := 0; i < len scr; i++) {

	# on commence, nous prenons la couleur
	if (n1 == 0)
	    col1 = scr[i];
	
	# si la couleur est la meme que le bg suppose
	if (scr[i] == col1) {
	    n1++;
	    continue;
	}

	# on a deux couleurs : c'est mal parti pour le mode solid
	if (n2 == 0) {
	    solid=0;
	    col2 = scr[i];
	}
	
	# on a toujours la couleur fg
	if (scr[i] == col2) {
	    n2++;
	    continue;
	}

	# on arrive ici si c'est ni mono et encore moins solid
	mono=0;
	# on peut arreter les frais
	break;
    }

    # Fixe les couleurs de fond et de premier plan
    if (n1>n2) {
	bg=col1;
	fg=col2;
    }
    else {
	bg=col2;
	fg=col1;
    }

    
    if (solid!=0)
	mode=rfb->hextileModeSolid;
    else if (mono!=0)
	mode=rfb->hextileModeMono;
    else
	# fg et bg ne seront pas pris en compte par l'appelant
	mode=rfb->hextileModeNone;

    #sys->print("Mode %d sur %s \n",mode,printrect(r));
    return(mode, bg, fg);
}
 
