#PROJECT : VNC 3.3 client and server for Inferno
#AUTHOR :  Marc Hufschmitt, marcas1@mime.univ-paris8.fr

# Fonctions pour le protocole rfb de VNC

implement RFBprotocol;

include "sys.m";
 sys:Sys;

include "rfbproto.m";

include "draw.m";
draw : Draw;
#Display : import draw;

# test Codage bigEndian pour les pixels
# Inferno est code en little endian (LSB ... MSB)
# variable privee


Send_ChallengeKey(c : ref Sys->Connection, key : array of byte) :int
{
    if (Send_Data(c, key) < 0) {
	return -1;
    }
    return 0;
} 

Send_AuthStatus(c : ref Sys->Connection, stat :int): int
{
    as := CARD(32,stat);
    if (Send_Data(c, as) < 0) {
	return -1;
    }
    return 0;
} 
Send_ConnectionFailed(c : ref Sys->Connection, m : int) : int
{
    return Send_Data(c,CARD(32,m));
}

Send_ConnectionFailedReason(c : ref Sys->Connection, r : string) : int
{
    buf := array[4] of byte;
    buf[0:] = CARD(32,len r);
    if (Send_Data(c, buf)<0)
	return 0;
    if (Send_Data(c,array of byte r)<0)
	return 0;
    return 1;
}

Receive_ResponseKey(c : ref Sys->Connection): array of byte
{
    key :=array[16] of byte;
    if (Read_Data(c, key) < 0) {
	return key[0:0];
    }
    return key;
} 
    
Send_ClientCutText(c : ref Sys->Connection, pbd : string) : int
{
    cct := array[8] of byte;
    cct[0:] = CARD(8,mtClientCutText);
    cct[4:] = CARD(32,len pbd);
    if (Send_Data(c, cct) < 0) {
	error("Client CutText entete"); 
	return -1;
    }
    if (Send_Data(c, array of byte pbd) < 0) {
	error("Client CutText message"); 
	return -1;
    }

    return 0;
    

}
Send_FramebufferUpdateRequest(c : ref Sys->Connection, inc : int , rect : Rectangle) : int
{
    fbur :=array[10] of byte;
    fbur[0:]=CARD(8,mtFramebufferUpdateRequest);

    fbur[1:]=CARD(8,inc);

    fbur[2:]=CARD(16,rect.min.x);
    fbur[4:]=CARD(16,rect.min.y);
    fbur[6:]=CARD(16,rect.dx());
    fbur[8:]=CARD(16,rect.dy());

    if (Send_Data(c, fbur) < 0) {
	error("FrameBuffer Update Request"); 
	return -1;
    }
    return 0;
}

Send_PointerEvent(c : ref Sys->Connection, x,y,b : int) : int
{
    pem := array [6] of byte;
    pem[0]  = byte PointerEvent;
    pem[1:] = CARD(8,b);
    pem[2:] = CARD(16,x);
    pem[4:] = CARD(16,y);
    if (WriteExact(c, pem, len pem) < 0) {
	error("Ecriture evenement pointeur"); 
	return -1;
    }
    return 0;
    
}

Send_KeyEvent (c : ref Sys->Connection, k,d : int) : int
{
    kem := array[8] of byte;
    kem[0]  = byte KeyEvent;
    kem[1:] = CARD(8,d);
    kem[4:] = CARD(32,k);
    if (WriteExact(c, kem, len kem) < 0) {
	error("Ecriture evenement clavier"); 
	return -1;
    }
    return 0;
    
}

VersionMsg.init (msg : self ref VersionMsg) 
{
    msg.struct=array of byte msg.VersionFormat;
}

AuthMsg.init(m: self ref AuthMsg, auth:int)
{
    m.struct = array[m.SIZE] of byte;
    m.struct[0:]=CARD(32,auth);
    m.keys=array[16] of byte;
    m.password = array[8] of { * => byte 0};
}

VersionMsg.cmp(m1:self ref VersionMsg, m2 : ref VersionMsg) : (int,int)
{
    major1:= int (string m1.struct[4:7]);
    minor1:= int (string m1.struct[8:11]);
    major2:= int (string m2.struct[4:7]);
    minor2:= int (string m2.struct[8:11]);
      return (major1-major2, minor1 - minor2);
}
 
ClientInitialisation.read(c : ref Sys->Connection) : int
{
    cim := array[1] of byte;
    if (ReadExact(c, cim ,1)<0) {
	error("Lecture shared flag"); 
	return -1;
    }
    return INT(cim);
}

ClientInitialisation.write(c : ref Sys->Connection, shared : int) : int
{
    cim :=CARD(8,shared);
    # shared flag
    if (WriteExact(c, cim, len cim) < 0) {
	error("Ecriture shared flag"); 
	return -1;
    }
    return 0;
}

ServerInitialisation.write(c : ref Sys->Connection, l, h : int, p : array of byte, sname : string) :int
{
    sim := array[ServerInitialisation.SIZE+PixelFormat.SIZE+len sname] of byte;
    sim[0:]=CARD(16,l);
    sim[2:]=CARD(16,h);
    sim[20:]=CARD(32,len sname);
    sim[4:]=p;
    sim[24:]=array of byte sname;
    if (WriteExact(c, sim , len sim)<0) {
	error("Lecture server Init"); 
	return -1;
    }
    return 0;
}

ServerInitialisation.read(c : ref Sys->Connection) : ( int , int, array of int, string)
{
    sim :=array[ServerInitialisation.SIZE+PixelFormat.SIZE] of byte;
    if (ReadExact(c, sim , len sim)<0) {
	error("Lecture server Init"); 
	return (-1,-1,nil,"");
    }
    l :=INT(sim[0:2]);
    h :=INT(sim[2:4]);
    lname :=INT(sim[20:24]);
    nbuf :=array[lname] of byte;
    if (ReadExact(c, nbuf , len nbuf)<0) {
	error("Lecture nom serveur"); 
	return (-1,-1,nil,"");
    }
    sname :=string nbuf[0:lname];
    pix := PixelFormat.get(sim[4:20]);
    return (l,h,pix,sname);
}



PixelFormat.truecolor (pf : self ref PixelFormat) : int
{
    return INT(pf.PIXEL_FORMAT[3:4]);
}

PixelFormat.set(pf : self ref PixelFormat, v : array of int) : array of byte
{
    pf.PIXEL_FORMAT = array[PixelFormat.SIZE] of byte;
    if (len v!=10) {
	error("Pixelformat parametres errones");
	pf.PIXEL_FORMAT=nil;
    }
    else 
    {
	pf.PIXEL_FORMAT[0:]=CARD(8,v[0]);
	pf.PIXEL_FORMAT[1:]=CARD(8,v[1]);
	pf.PIXEL_FORMAT[2:]=CARD(8,v[2]);
	pf.PIXEL_FORMAT[3:]=CARD(8,v[3]);
	pf.PIXEL_FORMAT[4:]=CARD(16,v[4]);
	pf.PIXEL_FORMAT[6:]=CARD(16,v[5]);
	pf.PIXEL_FORMAT[8:]=CARD(16,v[6]);
	pf.PIXEL_FORMAT[10:]=CARD(8,v[7]);
	pf.PIXEL_FORMAT[11:]=CARD(8,v[8]);
	pf.PIXEL_FORMAT[12:]=CARD(8,v[9]);
    }
    return pf.PIXEL_FORMAT;
}

PixelFormat.sprint(pixf : self ref PixelFormat) : string
{
    s : string;
    bpp := INT(pixf.PIXEL_FORMAT[0:1]);
    s  = "bpp :"+string bpp+" ";
    s += "depth :"+string INT(pixf.PIXEL_FORMAT[1:2])+" ";
    if (INT(pixf.PIXEL_FORMAT[2:3])==0)
	s += "little-endian ";
    else
	s += "big-endian ";
    if (INT(pixf.PIXEL_FORMAT[3:4])==0)
	s+="colormap";
    else {
	composantes := array[bpp] of byte;
	for (b:=bpp-1; b>=0; b--) {
	    composantes[b]=byte '.';
	    for (c:=0; c<3; c++) {
		i :=1<<b;  
		i &= INT(pixf.PIXEL_FORMAT[4+c*2:6+c*2])<<INT(pixf.PIXEL_FORMAT[10+c:11+c]);
		if (i) {
		    if (composantes[b]!=byte '.')
			composantes[b]=byte '!';
		    else
			case c {
			    0 => composantes[b]=byte 'R';
			    1 => composantes[b]=byte 'G';
			    2 => composantes[b]=byte 'B';
			}
		}
	    }
	}
	s+=string composantes;
    }
    return s;
}



PixelFormat.liste (pf : array of byte) : (int,int,int,int,int,int,int,int,int,int)#(bpp,depth,bigendian,truecol,rmask,gmask,bmask,rshift,gshift,bshift : int )
{
    return (INT(pf[0:1]),INT(pf[1:2]),INT(pf[2:3]),INT(pf[3:4]),
	    INT(pf[4:6]),INT(pf[6:8]),INT(pf[8:10]),
	    INT(pf[10:11]),INT(pf[11:12]),INT(pf[12:13]));
}

PixelFormat.get (pf : array of byte) : array of int
{
    v :=array[10] of int;
    v[0]=INT(pf[0:1]);
    v[1]=INT(pf[1:2]);
    v[2]=INT(pf[2:3]);
    v[3]=INT(pf[3:4]);
    v[4]=INT(pf[4:6]);
    v[5]=INT(pf[6:8]);
    v[6]=INT(pf[8:10]);
    v[7]=INT(pf[10:11]);
    v[8]=INT(pf[11:12]);
    v[9]=INT(pf[12:13]);
    return v;
}
PixelFormat.BGR233 () : array of int
{
    # BBGGGRRR
    BGR233format :=array[10] of 
    { 8, 8, 1, 1,
      7,       7,       3, 
      0,       3,       6,
      };
    return BGR233format;
}
    
PixelFormat.BGR466 () : array of int
{
    # BBBBGGGG GGRRRRRR
    BGR466format := array[10] of 
    { 16, 8, 1, 1,
      63,      63,      15, 
      0,       6,       12 
      };
    return BGR466format;
}

PixelFormat.colormap () : array of int
{
    # BBBBGGGG GGRRRRRR
    colormapformat := array[10] of 
    { 8, 8, 0, 0,
      0,      0,      0,
      0,      0,      0 
      };
    return colormapformat;
}



ReadExact(conn : ref sys->Connection, buf : array of byte, lon : int) : int 
{
  n: int;
    while (lon> 0) {
	n = sys->read(conn.dfd, buf[len buf-lon:], lon);
	if (n > 0) {
	    if (DEBUG)
		sys->print("Lecture '%s' [%d]\n",hexstrbuf( buf[0:]), len buf);
	    
	    lon -= n;
	} else {
	    return n;
	}
    }
    return 1;
}

Send_Data(conn: ref sys->Connection, buf: array of byte) : int
{
    written :=0;
    n : int;
    lon := len buf;
    
    if ((n=sys->write(conn.dfd, buf, lon))<len buf)
	return -1;
    if (DEBUG)
	sys->print("Ecriture '%s' [%d]\n",hexstrbuf (buf[0:n]), n);
    written+=n;
    return written;
}

Read_Data(conn: ref sys->Connection, buf: array of byte) : int
{
  read :=0;
  if ((n:= sys->read(conn.dfd, buf[read:], len buf))>0) {
      if (DEBUG)
	  sys->print("Lecture '%s' [%d]\n",hexstrbuf( buf[read:n]), n);
      read += n;
  }
  else 
      return -1;
  return read;
}

    
WriteExact(conn: ref sys->Connection, buf: array of byte, lon : int) : int
{
    n:int;
    ecrits:=0;
#    if (sys==nil) sys = load Sys Sys->PATH;

    while (ecrits < lon) {
	n = sys->write(conn.dfd, buf, lon);
		
	if (n > 0) {
	    if (DEBUG)
		sys->print("Ecriture '%s' [%d]\n",hexstrbuf (buf[0:n]), n);
	    buf = buf[n:];
	    ecrits += n;
	} else if (n == 0) {
	    exit;
        } else {
            return n;
	}
    }
      return 1;
}

hexstrbuf( b : array of byte) : string
{
    a : string="";
    # taille limite a 32 octets affichables
    for ( i:=0 ; i < len b && i< 32; i++) {
	a += sys->sprint("%2x",int b[i]);
	if (i<len b -1) a+=" ";
    }
    return a;
}


CARD(s : int, l :int) : array of byte
{
    ab := array[s>>3] of byte;
    #sys->print("CARD %d %d %d ",s,l,len ab);
    for (i:=0; i<len ab; i++) {
#	if (bigEndianPixFlag)
	    ab[i] = byte ((l>>(8*(len ab-i-1)) & 255));
#	else
#	    ab[i] = byte ((l>>(8*i) & 255));
    }
    #sys->print(" %d OK\n",len ab);
    return ab;
}

CARDRECT (r : Rectangle) : array of byte
{
    ra :=array[8] of byte;
    ra[0:]=CARD(16,r.min.x);
    ra[2:]=CARD(16,r.min.y);
    ra[4:]=CARD(16,r.dx());
    ra[6:]=CARD(16,r.dy());
    return ra;
}


INT(ab : array of byte) : int
{
  e:=0;
   for (i:=0; i<len ab; i++) {
#       if (bigEndianPixFlag!=0)
	    e+=(int ab[i])*(1<<(8*(len ab-i-1)));
#	else
#	    e+=(int ab[i])*(1<<(8*i));
    }
    return e;
}

init()
{
    errormessage="";
    sys = load Sys Sys->PATH;
    if (sys==nil)
	error("Chargement Module Sys");
    bigEndianPixFlag = 0;
}






error(s : string)
{
    errormessage= s;
    #sys->fprint(sys->fildes(2),"%s:%s -> %r\n",NAME,s);
}

lasterror() : string
{
    return errormessage;
}

Rectangle.ORrect(r : self Rectangle, s : Rectangle) : Rectangle
{
    t := r;
    if (r.min.x>s.min.x) t.min.x = s.min.x;
    if (r.min.y>s.min.y) t.min.y = s.min.y;
    if (r.max.x<s.max.x) t.max.x = s.max.x;
    if (r.max.y<s.max.y) t.max.y = s.max.y;
    return t;
}

Rectangle.dx( r : self Rectangle) : int
{
    return abs(r.max.x-r.min.x);
}

Rectangle.dy( r : self Rectangle) : int
{
    return abs(r.max.y-r.min.y);
}

Rectangle.surf(r : self Rectangle) : int
{
    return abs((r.max.x-r.min.x)*(r.max.y-r.min.y));
}

abs(x : int) : int
{
    if (x<0) return -x;
    return x;
}

BytePerPixelFromPixelFormat(pf :int ) : int
{
    case pf {
	PixFormat24bits => return 4;
	PixFormat16bits => return 2;
	PixFormatBGR233 or PixFormatCMap => return 1;
	* => exit;
    }
}

Encodages.accepte (e : self ref Encodages, enc :int ) : int
{
    for (i:=0; i< len e.tab; i++) {
	if (e.tab[i] == enc)
	    return (1==1);
    }
    return (0==1);
}

Encodages.rang (e : self ref Encodages, enc : int) : int
{
    for (i:=0; i< len e.tab; i++) {
	if (e.tab[i] == enc)
	    return i;
    }
    return -1;
}


Encodages.read (e : self ref Encodages, c : ref Sys->Connection,n : int) : int
{
    t := array[n*4] of byte;
    e.tab = array[n] of int;
    if (Read_Data(c,t)<0)
	return -1;
    for (i:=0; i< n; i++) {
	e.tab[i] = INT(t[0+i*4:4+i*4]);
	sys->print("Encoding %d\n",e.tab[i] );
    }
    return 0;
}

Encodages.write (e : self ref Encodages, c : ref Sys->Connection)
{
}
