#PROJECT : VNC 3.3 client and server for Inferno
#AUTHOR :  Marc Hufschmitt, marcas1@mime.univ-paris8.fr

# Definitions pour le protocole rfb de VNC

RFBprotocol: module
{
    
# ##############################################
# Constantes
# ##############################################
    
    DEBUG : int;    
    PATH : con "rfbproto.dis";
    NAME : con "RFB";

    errormessage : string;

    bigEndianPixFlag : int;

# taille Message d'init
    sz_rfbClientInitMsg : con 1;


    rfbClientInitMsg : byte;

# Serveur vers client
    mtFramebufferUpdate, mtSetColourMapEntries, Bell, ServerCutText: con iota;


# Client vers serveur
  SetPixelFormat: con 0;
  FixColourMapEntries: con 1;	# not currently supported 
  SetEncodings: con 2;
  mtFramebufferUpdateRequest: con  3;
    KeyEvent : con 4;
    KeyDown : con 1;
    KeyUp : con 0;
  PointerEvent: con 5;
  mtClientCutText: con 6;

# modes d'encodage
    Encodingraw : con 0;
    Encodingcopyrect : con 1;
    EncodingRRE : con 2;
    EncodingCoRRE : con 4;
    Encodinghextile : con 5;
    
    # hextile sub-encoding
    hextileRaw, 
    hextileBackgroundSpecified, 
    hextileForegroundSpecified, 
    hextileAnySubrects, 
    hextileSubrectsColoured : con 1<<iota; # valeurs 1 2 4 8 16
    
    hextileModeNone, 
    hextileModeMono, 
    hextileModeSolid : con iota; # valeurs 0 1 2
    
    Encodages : adt {
	tab : array of int;
	accepte : fn (e : self ref Encodages, enc  :int ) : int;
	rang : fn (e : self ref Encodages, enc : int) : int;
	read : fn (e : self ref Encodages, c : ref Sys->Connection, n : int) :int;
	write : fn (e : self ref Encodages, c : ref Sys->Connection);
    };

    Point : adt {
	x, y : int;
    };

    Pave : adt {
	couleur : byte;
	encoding : int;
	subencoding : int;
	r : Rectangle;
    };

    Rectangle : adt {
	#couleur : int;
	min, max : Point;
	dx : fn (r : self Rectangle) : int;
	dy : fn (r : self Rectangle) : int;
	ORrect : fn(r : self Rectangle, s : Rectangle) : Rectangle;
	surf : fn (r : self Rectangle) : int;
    };
 
    VersionMsg : adt {
      struct: array of byte;
      MajorVersion: con "003";
      MinorVersion: con "003";
      VersionFormat: con "RFB "+MajorVersion+"."+MinorVersion+"\n";
      SIZE: con 12;
      init:fn(msg : self ref VersionMsg);
	# Compare deux versions du protocole retourne msg1-msg2
	# la version mineure est la partie decimale
      cmp:fn(msg1:self ref VersionMsg, msg2: ref VersionMsg): (int,int);
    };
    
    ReadExact :fn(conn : ref sys->Connection, buf : array of byte, lon : int) : int;
    WriteExact :fn(conn: ref sys->Connection, buf: array of byte, lon : int) : int;
    Send_Data : fn (conn: ref sys->Connection, buf: array of byte) : int;
    Read_Data :fn (conn: ref sys->Connection, buf: array of byte) : int;
    INT : fn(ab : array of byte) : int;
    
    PixFormatNone, PixFormat8bits, PixFormat16bits, PixFormat24bits,
    PixFormatBGR233, PixFormatCMap : con iota;

    

    PixelFormat : adt {
	SIZE : con 16;
	numero : int;
	enc : ref Encodages;
	PIXEL_FORMAT : array of byte;
	set : fn(pf : self ref PixelFormat, v : array of int) : array of byte;
	truecolor : fn (pf : self ref PixelFormat): int;
	#bpp, depth, bigendian, truecolor,
	#rmax, gmax, bmax, rsh, gsh, bsh : int
	get : fn (pf : array of byte) : array of int;
	sprint : fn (pixf : self ref PixelFormat) : string;

	liste : fn (pf : array of byte) : 
	    (int,int,int,int,int,int,int,int,int,int);
	    #(bpp,depth,bigendian,truecol,rmask,gmask,bmask,rshift,gshift,bshift : int);
	#write : fn(pf : self ref PixelFormat, c: ref Sys->Connection);
	#read : fn(pf : self ref PixelFormat, c: ref Sys->Connection);
	BGR233 : fn () : array of int;
	BGR466 : fn () : array of int;
	colormap : fn() : array of int;
#  	# BBGGGRRR
    };

    ServerInitialisation : adt {
	write : fn(c : ref Sys->Connection, l :int , h : int, p : array of byte, sname : string) : int ;
	read : fn(c : ref Sys->Connection): ( int , int, array of int, string);
	SIZE : con 8;
    };

    ClientInitialisation : adt {
	write :fn(c : ref Sys->Connection, shared : int) : int;
	read :fn(c : ref Sys->Connection) : int;
	Shared : con 1;
	NoShare : con 0;
    };

    #rfbProtocolVersionFormat : con "RFB %s.%s\n";
    #rfbProtocolMajorVersion : con "003";
    #rfbProtocolMinorVersion : con "003";
    #rfbProtocolVersionMsg : string ;	# allow extra byte fo
    #sz_rfbProtocolVersionMsg : con 12;

    AuthMsg : adt {
	SIZE : con 4;
      struct:array of byte;
      keys:array of byte;
      password:array of byte;
      init:fn(m:self ref AuthMsg,auth:int);
    };

# Type d'authenfication
    ConnectionFailed :con 0;
# Pas d'authentification
    NoAuth :con 1;
# auth type VNC
    VNCAuth :con 2;
    
    # echec de l'auth
    AuthOK :con 0;
    Authfailed :con 1;
    Authtoomany :con 2;
    
# initialise struct
  init:fn();

    hexstrbuf :fn( b : array of byte) : string;
    CARD  : fn(s : int, l :int) : array of byte;
    CARDRECT : fn (r : Rectangle) : array of byte;
    Send_PointerEvent : fn(c : ref Sys->Connection, x,y,b : int) : int;
    Send_KeyEvent : fn(c : ref Sys->Connection, d,k : int) : int;
    Send_FramebufferUpdateRequest : fn(c : ref Sys->Connection,inc : int , r : Rectangle) : int;
    Send_ClientCutText : fn(c : ref Sys->Connection, pbd : string) : int;
    Send_ChallengeKey : fn(c : ref Sys->Connection, key : array of byte) :int;
    Receive_ResponseKey : fn(c : ref Sys->Connection): array of byte;
    Send_AuthStatus : fn(c : ref Sys->Connection, stat : int): int;
    Send_ConnectionFailedReason : fn(c : ref Sys->Connection, reason : string) : int;
    Send_ConnectionFailed : fn(c : ref Sys->Connection, mode : int) : int;
    lasterror : fn() : string;
    #ColorFromPixelFormat : fn(disp : ref Draw->Display , pb : array of byte, pxf : int, cmap : array of int,bigendian : int) :  byte;
    #ImageColorFromPixelFormat : fn(disp : ref Draw->Display , pb : array of byte, pxf : int, cmap : array of int,bigendian : int) : array of byte;
};
