#PROJECT : VNC 3.3 client and server for Inferno
#AUTHOR :  Marc Hufschmitt, marcas1@mime.univ-paris8.fr


#  Modified by Marc Hufschmitt, MIME Dept., University of Paris 8
#  VNC 3.3 for Lucent Technologies Inferno OS

#
# vncauth.b - Functions for VNC password management and authentication.
#

implement VNCauth;

include "sys.m";
sys : Sys;


include "vncauth.m";

# We use a fixed key to store passwords, since we assume that our local
# file system is secure but nonetheless don't want to store passwords
# as plaintext.
fixedkey := array[8] of { byte 23, byte 82, byte 107, byte 6, 
			  byte 35, byte 78, byte 88, byte 7};
DEBUG : int;
initdone :=0;

init()
{
    if (sys != nil)
	return;
    if ((sys = load Sys Sys->PATH) != nil)
	initdone = 1;
    else
	exit;
}
 
# Encrypt a password and store it in a file.
# Returns 0 if successful,
# -1 if the file could not be written.

EncryptAndStorePasswd(passwd : string, fname : string ) : int
{
    if (initdone == 0)
	init();

    fp : ref Sys->FD;
    i : int;
    encryptedPasswd := array[8] of { * => byte 0};

    if ((fp = sys->create(fname,sys->OWRITE,8r400 | 8r200)) == nil) 
	return -1;

    (ok,dir) := sys->fstat(fp);
    if (ok<0)
	return -1;
    dir.mode = 8r400 | 8r200; 
    #if ((sys->fwstat(fp, dir))<0)
    #return -1;
    

    # password is already padded with nulls 
    # just copy
    encryptedPasswd[0:] = array of byte passwd;

    # Do encryption in-place - this way we overwrite 
    # our copy of the plaintext password

    DESkey(fixedkey, DESencrypt);
    DEScode(encryptedPasswd, encryptedPasswd);

    sys->write(fp,encryptedPasswd,8);
  
    fp = nil;
    return 0;
}


# Decrypt a password from a file.  Returns a pointer to a newly allocated
# string containing the password or a null pointer if the password could
# not be retrieved for some reason.
#/

DecryptPasswdFromFile(fname : string) : string
{
    if (initdone == 0)
	init();

    fp : ref Sys->FD;
    i, ch : int;
    passwd := array[8] of byte;

    if ((fp = sys->open(fname,sys->OREAD)) == nil) 
	return "";

    if (sys->read(fp, passwd, len passwd)<=0)
	return "";
    fp = nil;
    
    DESkey(fixedkey, DESdecrypt);
    DEScode(passwd, passwd);
    
    return string passwd;
}


# Generate CHALLENGESIZE random bytes for use in challenge-response
# authentication.

randombuf(which, n: int): array of byte
{
    if (initdone == 0)
	init();
    
    buf := array[n] of byte;
    
    file : string;
    case(which){
	ReallyRandom =>
	file = "/dev/random";
	NotQuiteRandom =>
	file = "/dev/notquiterandom";
    }

    fd := sys->open(file, sys->OREAD);
    if(fd == nil){
	sys->print("can't open /dev/random\n");
	return buf;
    }
    sys->read(fd, buf, n);
    return buf;
}


# Encrypt CHALLENGESIZE bytes in memory using a password.

EncryptBytes(clef : array of byte, password : string)
{
    if (initdone == 0)
	init();

    key := array[8] of { * => byte 0};
    
    key[0:]=array of byte password;
    #DESKnL = array[32] of { * => 0};
    
    DESkey(key,DESencrypt);
    
    for (i := 0; i < CHALLENGESIZE; i += 8) {
	DEScode(clef[i:], clef[i:]);
    }

    # Clean memory
    key=array[8] of { * => byte 0};
    # and garbage
    key = nil;
}

DESkey(key:array of byte, edflag: int)
{
    pc1 := array[56] of {
      56, 48, 40, 32, 24, 16,  8,      0, 57, 49, 41, 33, 25, 17,
      9,  1, 58, 50, 42, 34, 26,     18, 10,  2, 59, 51, 43, 35,
      62, 54, 46, 38, 30, 22, 14,      6, 61, 53, 45, 37, 29, 21,
      13,  5, 60, 52, 44, 36, 28,     20, 12,  4, 27, 19, 11,  3 };
    
    pc2 := array[48] of {
        13, 16, 10, 23,  0,  4,  2, 27, 14,  5, 20,  9,
        22, 18, 11,  3, 25,  7, 15,  6, 26, 19, 12,  1,
        40, 51, 30, 36, 46, 54, 29, 39, 50, 44, 32, 47,
        43, 48, 38, 55, 33, 52, 45, 41, 49, 35, 28, 31 };

  
    bytebit := array[8] of {
	8r01, 8r02, 8r04, 8r010, 8r020, 8r040, 8r0100, 8r0200 };
    
    pc1m := array[56] of byte;

    i,j,l,m,n:int;

    bigbyte := array[24] of int;
    for (i=0; i<24; i++) {
	bigbyte[i] =  (1 << (23-i));
    }

    for ( j=0; j < 56; j++ ) {
	l = pc1[j]; # l contient des valeurs entre 0 et 62
	m = l & 07; # m contient les 3  LSbits de l
	if (((int key[(l >> 3)]) & bytebit[m]))
	    pc1m[j] = byte 1;
	else
	    pc1m[j] = byte 0;
	#sys->print("%d ",int pc1m[j]);
	#if (j==55) sys->print("\n");
    }
    
    totrot := array[16] of {
	1,2,4,6,8,10,12,14,15,17,19,21,23,25,27,28 };
    
    pcr :=array[56] of byte;
    kn :=array[32] of int;

    for( i=0; i < 16; i++ ) {
	if( edflag == DESdecrypt ) 
	    m = (15 - i) << 1;
	else 
	    m = i << 1;
	n = m + 1;
	kn[m] = kn[n] =  0;
	
	for( j = 0; j < 28; j++ ) {
	    l = j + totrot[i];
	    if( l < 28 ) 
		pcr[j] = pc1m[l];
	    else 
		pcr[j] = pc1m[l - 28];
	}
	
	for( j = 28; j < 56; j++ ) {
	    l = j + totrot[i];
	    if( l < 56 ) 
		pcr[j] = pc1m[l];
	    else 
		pcr[j] = pc1m[l - 28];
	}
	
	for( j = 0; j < 24; j++ ) {
	    if( int pcr[pc2[j]] ) 
		kn[m] |= bigbyte[j];
	    if( int pcr[pc2[j+24]] ) 
		kn[n] |= bigbyte[j];
	}
    }
    
    cookey(kn);
    
    return;
}


cookey(raw1 : array of int)
{
    praw0 :=0;
    praw1 :=0;
    pcook :=0;
    dough :=array[32] of int;
  
    for( i:= 0; i < 16; i++) {
	praw0 = praw1++;
	dough[pcook]    = ((raw1[praw0] & 16r00fc0000) << 6);
	#sys->print("%8ux ",dough[pcook]);
	dough[pcook]   |= ((raw1[praw0] & 16r00000fc0) << 10);
	#sys->print("%8ux ",dough[pcook]);
	dough[pcook]   |= ((raw1[praw1] & 16r00fc0000) >> 10);
	#sys->print("%8ux ",dough[pcook]);
	dough[pcook] |= ((raw1[praw1] & 16r00000fc0) >> 6);
	#sys->print("%8ux ",dough[pcook]);
	pcook++;
	
	dough[pcook]    = ((raw1[praw0] & 16r0003f000) << 12);
	dough[pcook]   |= ((raw1[praw0] & 16r0000003f) << 16);
	dough[pcook]   |= ((raw1[praw1] & 16r0003f000) >> 4);
	dough[pcook] |= ((raw1[praw1] & 16r0000003f));
	pcook++;
	praw1++;
    } 
    DESKnL = dough;
}

desfunc(block : array of int)
{
  work:int;
  pkeys:=0;

    leftt := block[0];
    right := block[1];

    work = ((leftt >> 4) ^ right) & 16r0f0f0f0f;
    right ^= work;
    leftt ^= (work << 4);
    work = ((leftt >> 16) ^ right) & 16r0000ffff;
    right ^= work;
    leftt ^= (work << 16);
    work = ((right >> 2) ^ leftt) & 16r33333333;
    leftt ^= work;
    right ^= (work << 2);
    work = ((right >> 8) ^ leftt) & 16r00ff00ff;
    leftt ^= work;
    right ^= (work << 8);
    right = ((right << 1) | ((right >> 31) &  1)) &  int 16rffffffff;
    work = (leftt ^ right) &  int 16raaaaaaaa;
    leftt ^= work;
    right ^= work;
    leftt = ((leftt << 1) | ((leftt >> 31) &  1)) &  int 16rffffffff;

    fval : int;
    for( round := 0; round < 8; round++ ) {
	work  = ((right << 28) | ((right >> 4) & 16r0fffffff));
	work ^= DESKnL[pkeys];
	pkeys++;
	fval  = SP[64*6 + int (work        & 16r3f)];
	fval |= SP[64*4 + int ((work >>  8) & 16r3f)];
	fval |= SP[64*2 + int ((work >> 16) & 16r3f)];
	fval |= SP[64*0 + int ((work >> 24) & 16r3f)];
	work  = right ^ DESKnL[pkeys];

	pkeys++;
	fval |= SP[64*7 + int ( work        & 16r3f)];
	fval |= SP[64*5 + int ((work >>  8) & 16r3f)];
	fval |= SP[64*3 + int ((work >> 16) & 16r3f)];
	fval |= SP[64*1 + int ((work >> 24) & 16r3f)];
	leftt ^= fval;

	work  = (leftt << 28) | ((leftt >> 4) & 16r0fffffff);
	work ^= DESKnL[pkeys];
	pkeys++;
	fval  = SP[64*6 + int ( work        & 16r3f)];
	fval |= SP[64*4 + int ((work >>  8) & 16r3f)];
	fval |= SP[64*2 + int ((work >> 16) & 16r3f)];
	fval |= SP[64*0 + int ((work >> 24) & 16r3f)];
	work  = leftt ^ DESKnL[pkeys];
	pkeys++;
	fval |= SP[64*7 + int ( work        & 16r3f)];
	fval |= SP[64*5 + int ((work >>  8) & 16r3f)];
	fval |= SP[64*3 + int ((work >> 16) & 16r3f)];
	fval |= SP[64*1 + int ((work >> 24) & 16r3f)];
	right ^= fval;

    }
  
    
    right = (right << 31) | ((right >> 1) & 16r7fffffff);
    work = (leftt ^ right) &  int 16raaaaaaaa;
    leftt ^= work;
    right ^= work;
    leftt = (leftt << 31) | ((leftt >> 1) & 16r7fffffff);
    work = ((leftt >> 8) ^ right) &  16r00ff00ff;
    right ^= work;
    leftt ^= (work << 8);
    work = ((leftt >> 2) ^ right) & 16r33333333;
    right ^= work;
    leftt ^= (work << 2);
    work = ((right >> 16) ^ leftt) & 16r0000ffff;
    leftt ^= work;
    right ^= (work << 16);
    work = ((right >> 4) ^ leftt) &  16r0f0f0f0f;
    leftt ^= work;
    right ^= (work << 4);
    block[0] = right;
    block[1] = leftt;
}

# Placer 8 octets dans un tableau de 2 entier de 32 bits
scrunch(from: array of byte, into:array of int)
{
    for (i:=0; i<4; i++) {
	into[0] <<=8;
	into[0] |= (int from[i]);
	into[1] <<=8;
	into[1] |= (int from[i+4]);
    }
    
}

# Placer 2 entiers 32 bits dans un tableau de 8 octets
unscrunch(out : array of int, into : array of byte)
{
    for (i:=0; i<8; i++) {
	into[i]= byte ((out[i/4] >> (24-8*(i%4))) & 255);
    }
}

# Encode la clef avec elle meme
DEScode(in:array of byte, out: array of byte)
{
    work := array[2] of int;
    #sys->print("On va scruncher in %s ->work!!!\n",hexstrbuf(in));
    scrunch(in,work);
    #sys->print("On va coder work %ux %ux !!!\n",work[0],work[1]);
    desfunc(work);
    #sys->print("On va unscruncher work %ux %ux !!!\n",work[0],work[1]);
    unscrunch(work,out);
    #sys->print("On a %s !!!\n",hexstrbuf(out));
    
}

    SP := array[512] of { 
	# **** SP1  ****
	
	int 16r01010400, int 16r00000000, int 16r00010000, int 16r01010404,
	int 16r01010004, int 16r00010404, int 16r00000004, int 16r00010000,
	int 16r00000400, int 16r01010400, int 16r01010404, int 16r00000400,
	int 16r01000404, int 16r01010004, int 16r01000000, int 16r00000004,
	int 16r00000404, int 16r01000400, int 16r01000400, int 16r00010400,
	int 16r00010400, int 16r01010000, int 16r01010000, int 16r01000404,
	int 16r00010004, int 16r01000004, int 16r01000004, int 16r00010004,
     int 16r00000000, int 16r00000404, int 16r00010404, int 16r01000000,
     int 16r00010000, int 16r01010404, int 16r00000004, int 16r01010000,
     int 16r01010400, int 16r01000000, int 16r01000000, int 16r00000400,
     int 16r01010004, int 16r00010000, int 16r00010400, int 16r01000004,
     int 16r00000400, int 16r00000004, int 16r01000404, int 16r00010404,
     int 16r01010404, int 16r00010004, int 16r01010000, int 16r01000404,
     int 16r01000004, int 16r00000404, int 16r00010404, int 16r01010400,
     int 16r00000404, int 16r01000400, int 16r01000400, int 16r00000000,
     int 16r00010004, int 16r00010400, int 16r00000000, int 16r01010004, 

     int 16r80108020, int 16r80008000, int 16r00008000, int 16r00108020,
     int 16r00100000, int 16r00000020, int 16r80100020, int 16r80008020,
     int 16r80000020, int 16r80108020, int 16r80108000, int 16r80000000,
     int 16r80008000, int 16r00100000, int 16r00000020, int 16r80100020,
     int 16r00108000, int 16r00100020, int 16r80008020, int 16r00000000,
     int 16r80000000, int 16r00008000, int 16r00108020, int 16r80100000,
     int 16r00100020, int 16r80000020, int 16r00000000, int 16r00108000,
     int 16r00008020, int 16r80108000, int 16r80100000, int 16r00008020,
     int 16r00000000, int 16r00108020, int 16r80100020, int 16r00100000,
     int 16r80008020, int 16r80100000, int 16r80108000, int 16r00008000,
     int 16r80100000, int 16r80008000, int 16r00000020, int 16r80108020,
     int 16r00108020, int 16r00000020, int 16r00008000, int 16r80000000,
     int 16r00008020, int 16r80108000, int 16r00100000, int 16r80000020,
     int 16r00100020, int 16r80008020, int 16r80000020, int 16r00100020,
     int 16r00108000, int 16r00000000, int 16r80008000, int 16r00008020,
     int 16r80000000, int 16r80100020, int 16r80108020, int 16r00108000, 


     int 16r00000208, int 16r08020200, int 16r00000000, int 16r08020008,
     int 16r08000200, int 16r00000000, int 16r00020208, int 16r08000200,
     int 16r00020008, int 16r08000008, int 16r08000008, int 16r00020000,
     int 16r08020208, int 16r00020008, int 16r08020000, int 16r00000208,
     int 16r08000000, int 16r00000008, int 16r08020200, int 16r00000200,
     int 16r00020200, int 16r08020000, int 16r08020008, int 16r00020208,
     int 16r08000208, int 16r00020200, int 16r00020000, int 16r08000208,
     int 16r00000008, int 16r08020208, int 16r00000200, int 16r08000000,
     int 16r08020200, int 16r08000000, int 16r00020008, int 16r00000208,
     int 16r00020000, int 16r08020200, int 16r08000200, int 16r00000000,
     int 16r00000200, int 16r00020008, int 16r08020208, int 16r08000200,
     int 16r08000008, int 16r00000200, int 16r00000000, int 16r08020008,
     int 16r08000208, int 16r00020000, int 16r08000000, int 16r08020208,
     int 16r00000008, int 16r00020208, int 16r00020200, int 16r08000008,
     int 16r08020000, int 16r08000208, int 16r00000208, int 16r08020000,
     int 16r00020208, int 16r00000008, int 16r08020008, int 16r00020200, 


     int 16r00802001, int 16r00002081, int 16r00002081, int 16r00000080,
     int 16r00802080, int 16r00800081, int 16r00800001, int 16r00002001,
     int 16r00000000, int 16r00802000, int 16r00802000, int 16r00802081,
     int 16r00000081, int 16r00000000, int 16r00800080, int 16r00800001,
     int 16r00000001, int 16r00002000, int 16r00800000, int 16r00802001,
     int 16r00000080, int 16r00800000, int 16r00002001, int 16r00002080,
     int 16r00800081, int 16r00000001, int 16r00002080, int 16r00800080,
     int 16r00002000, int 16r00802080, int 16r00802081, int 16r00000081,
     int 16r00800080, int 16r00800001, int 16r00802000, int 16r00802081,
     int 16r00000081, int 16r00000000, int 16r00000000, int 16r00802000,
     int 16r00002080, int 16r00800080, int 16r00800081, int 16r00000001,
     int 16r00802001, int 16r00002081, int 16r00002081, int 16r00000080,
     int 16r00802081, int 16r00000081, int 16r00000001, int 16r00002000,
     int 16r00800001, int 16r00002001, int 16r00802080, int 16r00800081,
     int 16r00002001, int 16r00002080, int 16r00800000, int 16r00802001,
     int 16r00000080, int 16r00800000, int 16r00002000, int 16r00802080, 


     int 16r00000100, int 16r02080100, int 16r02080000, int 16r42000100,
     int 16r00080000, int 16r00000100, int 16r40000000, int 16r02080000,
     int 16r40080100, int 16r00080000, int 16r02000100, int 16r40080100,
     int 16r42000100, int 16r42080000, int 16r00080100, int 16r40000000,
     int 16r02000000, int 16r40080000, int 16r40080000, int 16r00000000,
     int 16r40000100, int 16r42080100, int 16r42080100, int 16r02000100,
     int 16r42080000, int 16r40000100, int 16r00000000, int 16r42000000,
     int 16r02080100, int 16r02000000, int 16r42000000, int 16r00080100,
     int 16r00080000, int 16r42000100, int 16r00000100, int 16r02000000,
     int 16r40000000, int 16r02080000, int 16r42000100, int 16r40080100,
     int 16r02000100, int 16r40000000, int 16r42080000, int 16r02080100,
     int 16r40080100, int 16r00000100, int 16r02000000, int 16r42080000,
     int 16r42080100, int 16r00080100, int 16r42000000, int 16r42080100,
     int 16r02080000, int 16r00000000, int 16r40080000, int 16r42000000,
     int 16r00080100, int 16r02000100, int 16r40000100, int 16r00080000,
     int 16r00000000, int 16r40080000, int 16r02080100, int 16r40000100, 


     int 16r20000010, int 16r20400000, int 16r00004000, int 16r20404010,
     int 16r20400000, int 16r00000010, int 16r20404010, int 16r00400000,
     int 16r20004000, int 16r00404010, int 16r00400000, int 16r20000010,
     int 16r00400010, int 16r20004000, int 16r20000000, int 16r00004010,
     int 16r00000000, int 16r00400010, int 16r20004010, int 16r00004000,
     int 16r00404000, int 16r20004010, int 16r00000010, int 16r20400010,
     int 16r20400010, int 16r00000000, int 16r00404010, int 16r20404000,
     int 16r00004010, int 16r00404000, int 16r20404000, int 16r20000000,
     int 16r20004000, int 16r00000010, int 16r20400010, int 16r00404000,
     int 16r20404010, int 16r00400000, int 16r00004010, int 16r20000010,
     int 16r00400000, int 16r20004000, int 16r20000000, int 16r00004010,
     int 16r20000010, int 16r20404010, int 16r00404000, int 16r20400000,
     int 16r00404010, int 16r20404000, int 16r00000000, int 16r20400010,
     int 16r00000010, int 16r00004000, int 16r20400000, int 16r00404010,
     int 16r00004000, int 16r00400010, int 16r20004010, int 16r00000000,
     int 16r20404000, int 16r20000000, int 16r00400010, int 16r20004010, 


     int 16r00200000, int 16r04200002, int 16r04000802, int 16r00000000,
     int 16r00000800, int 16r04000802, int 16r00200802, int 16r04200800,
     int 16r04200802, int 16r00200000, int 16r00000000, int 16r04000002,
     int 16r00000002, int 16r04000000, int 16r04200002, int 16r00000802,
     int 16r04000800, int 16r00200802, int 16r00200002, int 16r04000800,
     int 16r04000002, int 16r04200000, int 16r04200800, int 16r00200002,
     int 16r04200000, int 16r00000800, int 16r00000802, int 16r04200802,
     int 16r00200800, int 16r00000002, int 16r04000000, int 16r00200800,
     int 16r04000000, int 16r00200800, int 16r00200000, int 16r04000802,
     int 16r04000802, int 16r04200002, int 16r04200002, int 16r00000002,
     int 16r00200002, int 16r04000000, int 16r04000800, int 16r00200000,
     int 16r04200800, int 16r00000802, int 16r00200802, int 16r04200800,
     int 16r00000802, int 16r04000002, int 16r04200802, int 16r04200000,
     int 16r00200800, int 16r00000000, int 16r00000002, int 16r04200802,
     int 16r00000000, int 16r00200802, int 16r04200000, int 16r00000800,
     int 16r04000002, int 16r04000800, int 16r00000800, int 16r00200002, 


     int 16r10001040, int 16r00001000, int 16r00040000, int 16r10041040,
     int 16r10000000, int 16r10001040, int 16r00000040, int 16r10000000,
     int 16r00040040, int 16r10040000, int 16r10041040, int 16r00041000,
     int 16r10041000, int 16r00041040, int 16r00001000, int 16r00000040,
     int 16r10040000, int 16r10000040, int 16r10001000, int 16r00001040,
     int 16r00041000, int 16r00040040, int 16r10040040, int 16r10041000,
     int 16r00001040, int 16r00000000, int 16r00000000, int 16r10040040,
     int 16r10000040, int 16r10001000, int 16r00041040, int 16r00040000,
     int 16r00041040, int 16r00040000, int 16r10041000, int 16r00001000,
     int 16r00000040, int 16r10040040, int 16r00001000, int 16r00041040,
     int 16r10001000, int 16r00000040, int 16r10000040, int 16r10040000,
     int 16r10040040, int 16r10000000, int 16r00040000, int 16r10001040,
     int 16r00000000, int 16r10041040, int 16r00040040, int 16r10000040,
     int 16r10040000, int 16r10001000, int 16r10001040, int 16r00000000,
     int 16r10041040, int 16r00041000, int 16r00041000, int 16r00001040,
     int 16r00001040, int 16r00040040, int 16r10000000, int 16r10041000 
     };
